from typing import Dict
from typing import List
from typing import Optional

import attrs
import pendulum

from tecton._internals import errors
from tecton._internals import time_utils
from tecton._internals.data_frame_helper import get_time_limits_of_dataframe
from tecton._internals.utils import get_time_limits_of_pandas_dataframe
from tecton_core import conf
from tecton_core.compute_mode import ComputeMode
from tecton_core.pipeline_common import get_time_window_from_data_source_node
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_interface import QueryNode
from tecton_core.query.node_utils import get_first_input_node_of_class
from tecton_core.query.nodes import AddAnchorTimeNode
from tecton_core.query.nodes import AddDurationNode
from tecton_core.query.nodes import AddEffectiveTimestampNode
from tecton_core.query.nodes import AsofJoinFullAggNode
from tecton_core.query.nodes import AsofJoinNode
from tecton_core.query.nodes import ConvertTimestampToUTCNode
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.nodes import EntityFilterNode
from tecton_core.query.nodes import FeatureTimeFilterNode
from tecton_core.query.nodes import FeatureViewPipelineNode
from tecton_core.query.nodes import MockDataSourceScanNode
from tecton_core.query.nodes import OfflineStoreScanNode
from tecton_core.query.nodes import PartialAggNode
from tecton_core.query.nodes import RenameColsNode
from tecton_core.query.nodes import RespectFeatureStartTimeNode
from tecton_core.query.nodes import SelectDistinctNode
from tecton_core.query.nodes import StagingNode
from tecton_core.query.nodes import TextEmbeddingInferenceNode
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_core.query.rewrite import Rewrite


class MockDataRewrite(Rewrite):
    """
    Replace DataSourceScanNode with MockDataSourceScanNode based on DataSource ID map.
    """

    def __init__(self, mock_data: Dict[str, NodeRef]):
        self.mock_data = mock_data

    def rewrite(self, tree: NodeRef):
        if isinstance(tree.node, DataSourceScanNode):
            node = tree.node
            if node.ds.id in self.mock_data:
                # Replace with mock data
                tree.node = MockDataSourceScanNode(
                    node.dialect,
                    node.compute_mode,
                    self.mock_data[node.ds.id],
                    node.ds,
                    self.mock_data[node.ds.id].columns,
                    node.start_time,
                    node.end_time,
                )
        else:
            for i in tree.inputs:
                self.rewrite(i)


class SpineEntityPushdown(Rewrite):
    """Filters the original feature data with respect to the entities contained in a spine.

    This should be applied to AsofJoinNodes and AsofJoinFullAggNodes since both of have sorts (as part of windows), for
    which we would like to minimize memory usage.
    """

    def __init__(self):
        pass

    def rewrite(self, tree: NodeRef):
        if isinstance(tree.node, AsofJoinNode):
            self.rewrite_asof(tree)
        elif isinstance(tree.node, AsofJoinFullAggNode):
            self.rewrite_asof_full_agg(tree)
        else:
            for i in tree.inputs:
                self.rewrite(i)

    def rewrite_asof(self, tree: NodeRef):
        node = tree.node
        # In our current usage of AsofJoinNode, we always pass in the spine on the left.
        spine_df = node.left_container.node.deepcopy()
        self.pushdown_entities(node.right_container.node, spine_df, node.join_cols)

    def rewrite_asof_full_agg(self, tree: NodeRef):
        node = tree.node

        if not node.enable_spine_entity_pushdown_rewrite:
            return

        user_provided_spine_node_ref = get_first_input_node_of_class(node.spine, UserSpecifiedDataNode).as_ref()

        # Note: we can't simply using join keys to pushdown entity because wildcard FV can have a spine missing wildcard
        # key, so we need to check what exact entities are used in spine if wildcard key is present, and only push those
        # entities down.
        entities_to_push_down = node.fdw.join_keys
        if node.fdw.wildcard_join_key and node.fdw.wildcard_join_key not in user_provided_spine_node_ref.columns:
            entities_to_push_down = set(entities_to_push_down) - {node.fdw.wildcard_join_key}

        # entity_rewritable ensures there is a UserSpecifiedDataNode in the spine.
        self.pushdown_entities(
            node.partial_agg_node, user_provided_spine_node_ref.deepcopy(), list(entities_to_push_down)
        )

    def pushdown_entities(self, tree: NodeRef, spine: NodeRef, join_cols: List[str]):
        node = tree.node
        can_be_pushed_down = (
            RespectFeatureStartTimeNode,
            RenameColsNode,
            FeatureTimeFilterNode,
            AddAnchorTimeNode,
            AddDurationNode,
            AddEffectiveTimestampNode,
            PartialAggNode,
            StagingNode,
            ConvertTimestampToUTCNode,
            TextEmbeddingInferenceNode,
        )
        if isinstance(node, can_be_pushed_down):
            self.pushdown_entities(node.input_node, spine, join_cols)
        else:
            entities_node = SelectDistinctNode(node.dialect, node.compute_mode, spine, join_cols).as_ref()
            if isinstance(node, OfflineStoreScanNode):
                node = attrs.evolve(node, entity_filter=entities_node)

            tree.node = EntityFilterNode(node.dialect, node.compute_mode, node.as_ref(), entities_node, join_cols)


class SpineTimePushdown(Rewrite):
    """
    Adds time filter ranges based on the time limits of the spine.

    For example, if the spine in a fv.ghf() has time range [start, end], then the rewrite might apply that time range
    as a filter to any FeatureViewPipelineNodes, OfflineStoreScanNodes, or DataSourceScanNodes that are part of the
    feature data (with the appropriate modifications to the time range to account for factors such as batch schedule,
    data delay, etc.).
    """

    def __init__(self):
        # The spine time limits are the same for all spines used throughout the query, so we only calculate once.
        self.spine_time_limits: Optional[pendulum.Period] = None

    def rewrite(self, tree: NodeRef):
        if isinstance(tree.node, AsofJoinNode):
            self.rewrite_asof(tree)
        elif isinstance(tree.node, AsofJoinFullAggNode):
            self.rewrite_asof_full_agg(tree)
        else:
            for i in tree.inputs:
                self.rewrite(i)

    def rewrite_asof(self, tree: NodeRef):
        node = tree.node
        # In our current usage of the code, we always pass in the spine on the left side of the asof join.
        # This rewrite is still applicable for any type of dataframe on the left, but referring to it as spine
        # to make the naming match up with the aggregate case (in which the spine is not on the left).
        if self.spine_time_limits is None:
            cur_node = node.left_container.node.node
            self.spine_time_limits = _get_spine_time_limits(cur_node)

        self.pushdown_time_range(node.right_container.node, self.spine_time_limits)

    def rewrite_asof_full_agg(self, tree: NodeRef):
        """Computes the spine time limits and pushes them down to all relevant nodes in the partial aggregates."""
        node = tree.node

        if not node.enable_spine_time_pushdown_rewrite:
            return

        if self.spine_time_limits is None:
            self.spine_time_limits = _get_spine_time_limits(node)
        self.pushdown_time_range(node.partial_agg_node, self.spine_time_limits)

    # Push down and convert spine time filter to either raw data or feature time filter at the DataSourceScanNode or OfflineStoreScanNode.
    # Nodes that do not affect the correlation with the spine time range are enumerated in the can_be_pushed_down list.
    def pushdown_time_range(self, tree: NodeRef, spine_time_limits: pendulum.Period):
        node = tree.node
        can_be_pushed_down = (
            RespectFeatureStartTimeNode,
            RenameColsNode,
            PartialAggNode,
            FeatureTimeFilterNode,
            AddAnchorTimeNode,
            AddDurationNode,
            AddEffectiveTimestampNode,
            StagingNode,
            ConvertTimestampToUTCNode,
        )
        if isinstance(node, can_be_pushed_down):
            self.pushdown_time_range(node.input_node, spine_time_limits)
        elif isinstance(node, (OfflineStoreScanNode, FeatureViewPipelineNode)):
            feature_time_limits = time_utils.get_feature_data_time_limits(
                fd=node.feature_definition_wrapper, spine_time_limits=spine_time_limits
            )
            if isinstance(node, FeatureViewPipelineNode):
                # swap the pipeline node to add new time limits
                node = FeatureViewPipelineNode(
                    node.dialect,
                    node.compute_mode,
                    node.inputs_map,
                    node.feature_definition_wrapper,
                    feature_time_limits,
                    node.check_view_schema,
                )

                for n in node.inputs:
                    data_source_scan_node_ref = get_first_input_node_of_class(n, DataSourceScanNode, as_node_ref=True)
                    if data_source_scan_node_ref is None:
                        continue
                    data_source_scan_node = data_source_scan_node_ref.node
                    if data_source_scan_node is not None:
                        # this method will convert aligned_feature_time_limits to raw data time limits by accounting for FilteredSource offsets etc.
                        data_time_filter = get_time_window_from_data_source_node(
                            feature_time_limits,
                            node.feature_definition_wrapper.batch_materialization_schedule,
                            data_source_scan_node.ds_node,
                        )
                        if data_time_filter is not None:
                            data_source_scan_node_ref.node = attrs.evolve(
                                data_source_scan_node_ref.node,
                                start_time=data_time_filter.start,
                                end_time=data_time_filter.end,
                            )

                tree.node = FeatureTimeFilterNode(
                    node.dialect,
                    node.compute_mode,
                    ConvertTimestampToUTCNode.for_feature_definition(
                        node.dialect, node.compute_mode, node.feature_definition_wrapper, node.as_ref()
                    ),
                    feature_time_limits,
                    node.feature_definition_wrapper.time_range_policy,
                    node.feature_definition_wrapper.timestamp_key,
                    node.feature_definition_wrapper.timestamp_key,
                )
            elif isinstance(node, OfflineStoreScanNode):
                tree.node = attrs.evolve(node, partition_time_filter=feature_time_limits)


# Mutates the input
def rewrite_tree_for_spine(tree: NodeRef):
    if not conf.get_bool("QUERY_REWRITE_ENABLED"):
        return
    rewrites = [SpineTimePushdown(), SpineEntityPushdown()]
    for rewrite in rewrites:
        rewrite.rewrite(tree)


def _get_spine_time_limits(cur_node: QueryNode) -> pendulum.Period:
    user_specified_data_node = get_first_input_node_of_class(cur_node.as_ref(), UserSpecifiedDataNode)
    if not user_specified_data_node:
        # We don't expect this to occur
        msg = "Expected spine to contain a UserSpecifiedDataNode, but it did not."
        raise ValueError(msg)
    timestamp_key = user_specified_data_node.metadata["timestamp_key"]
    data = user_specified_data_node.data
    if cur_node.compute_mode == ComputeMode.SPARK:
        limits = get_time_limits_of_dataframe(data.to_spark(), timestamp_key)
    else:
        limits = get_time_limits_of_pandas_dataframe(data.to_pandas(), timestamp_key)
    if limits is None:
        raise errors.INVALID_NULL_SPINE_TIME_KEY
    return limits



import re
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy
import pandas
from pyspark.sql.types import ArrayType
from pyspark.sql.types import MapType
from pyspark.sql.types import StructType
from pyspark.sql.types import TimestampType

from tecton_core import conf
from tecton_core import specs
from tecton_core.compute_mode import ComputeMode
from tecton_core.errors import UDF_ERROR
from tecton_core.errors import UDF_TYPE_ERROR
from tecton_core.id_helper import IdHelper
from tecton_core.pipeline_common import constant_node_to_value
from tecton_core.query_consts import udf_internal
from tecton_core.realtime_context import RealtimeContext
from tecton_proto.args.pipeline__client_pb2 import Pipeline
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.args.pipeline__client_pb2 import TransformationNode
from tecton_proto.args.transformation__client_pb2 import TransformationMode
from tecton_spark import feature_view_spark_utils
from tecton_spark.spark_pipeline import SparkFeaturePipeline
from tecton_spark.type_annotations import PySparkDataFrame


# TODO(TEC-8978): remove \. from namespace regex when FWv3 FVs are no longer supported.
_NAMESPACE_SEPARATOR_REGEX = re.compile(r"__|\.")


def feature_name(namespaced_feature_name: str) -> str:
    """Gets the base feature name from a namespaced_feature_name (e.g. feature_view__feature)

    Supports both `__` (fwv5) and `.` (fwv3) separators. Does two attempts at
    getting the feature name since `__` was allowed in feature view names in
    fwv3.
    """

    spl = _NAMESPACE_SEPARATOR_REGEX.split(namespaced_feature_name)
    if len(spl) == 2:
        return spl[1]

    return namespaced_feature_name.split(".")[1]


# For Pandas-mode:
# A pandas udf takes as inputs List[pandas.Series] and outputs List[pandas.Series]
# However, the user-defined transforms take as input pandas.DataFrame and output
# pandas.DataFrame. RealtimeFeaturePipeline will construct a UDF Wrapper functions
# that translates the inputs and outputs and performs some type checking.
#
# The general idea is that each Node of the pipeline evaluates to a pandas.DataFrame.
# This is what we want since the user-defined transforms take pandas.DataFrame
# as inputs both from RequestDataSourceNode or FeatureViewNode.
# pandas_udf_wrapper then typechecks and translates the final pandas.DataFrame into a
# jsonized pandas.Series to match what spark expects.
#
# For Python-mode, we can use a simpler wrapper function for the udf because we don't do
# any spark<->pandas type conversions.
class RealtimeFeaturePipeline(SparkFeaturePipeline):
    _VALID_MODES = ["pipeline", "python", "pandas"]

    def __init__(
        self,
        name: str,
        pipeline: Pipeline,
        transformations: List[specs.TransformationSpec],
        # maps input + feature name to arg index that udf function wrapper will be called with.
        # this is needed because we need to know which pandas.Series that are inputs to this
        # function correspond to the desired request context fields or dependent fv features
        # that the customer-defined udf uses.
        udf_arg_idx_map: Dict[str, int],
        output_schema: Optional[StructType],
        events_df_timestamp_field: Optional[str] = None,
        # the id of this OnDemandFeatureView; only required to be set when reading from source data
        fv_id: Optional[str] = None,
        data_source_inputs: Optional[Dict[str, Union[Dict[str, Any], pandas.DataFrame, RealtimeContext]]] = None,
    ) -> None:
        self._pipeline = pipeline
        self._name = name
        self._fv_id = fv_id
        self.udf_arg_idx_map = udf_arg_idx_map
        self._id_to_transformation = {t.id: t for t in transformations}
        self._output_schema = output_schema
        self._data_source_inputs = data_source_inputs
        self._events_df_timestamp_field = events_df_timestamp_field
        # In Spark, the UDF cannot reference a proto enum, so instead save mode as a string
        self.mode = (
            "python"
            if self._id_to_transformation[
                IdHelper.to_string(self._pipeline.root.transformation_node.transformation_id)
            ].transformation_mode
            == TransformationMode.TRANSFORMATION_MODE_PYTHON
            else "pandas"
        )
        # Access this conf value outside of the UDF to avoid doing it many times and avoiding any worker/driver state issues.
        self._should_check_output_schema = conf.get_bool("TECTON_PYTHON_ODFV_OUTPUT_SCHEMA_CHECK_ENABLED")

    def is_pandas_mode(self):
        return self.mode == "pandas"

    def is_python_mode(self):
        return self.mode == "python"

    def get_dataframe(self) -> PySparkDataFrame:
        df = self._node_to_value(self._pipeline.root)
        return df

    def py_wrapper(self, *args):
        assert self.is_python_mode()
        self._udf_args: List = args
        res = self._node_to_value(self._pipeline.root)
        if self._should_check_output_schema:
            feature_view_spark_utils.check_python_odfv_output_schema(res, self._output_schema, self._name)
        return res

    def pandas_udf_wrapper(self, *args):
        assert self.is_pandas_mode()

        import json

        import pandas

        # self.udf_arg_idx_map tells us which of these pandas.Series correspond to a given
        # RequestDataSource or FeatureView input
        self._udf_args: List[pandas.Series] = args

        output_df = self._node_to_value(self._pipeline.root)

        assert isinstance(
            output_df, pandas.DataFrame
        ), f"Transformer returns {str(output_df)}, but must return a pandas.DataFrame instead."

        for field in self._output_schema:
            assert field.name in output_df.columns, (
                f"Expected output schema field '{field.name}' not found in columns of DataFrame returned by "
                f"'{self._name}': [" + ", ".join(output_df.columns) + "]"
            )
            if isinstance(field.dataType, (ArrayType, MapType, StructType, TimestampType)):
                output_df[field.name] = output_df[field.name].apply(
                    RealtimeFeaturePipeline._convert_object_to_serializable_format
                )

        output_strs = []

        # itertuples() is used instead of iterrows() to preserve type safety.
        # See notes in https://pandas.pydata.org/pandas-docs/version/0.17.1/generated/pandas.DataFrame.iterrows.html.
        for row in output_df.itertuples(index=False):
            output_strs.append(json.dumps(row._asdict()))
        return pandas.Series(output_strs)

    @staticmethod
    def _convert_object_to_serializable_format(item):
        if isinstance(item, numpy.ndarray):
            return [RealtimeFeaturePipeline._convert_object_to_serializable_format(value) for value in item.tolist()]
        elif isinstance(item, dict):
            return {
                key: RealtimeFeaturePipeline._convert_object_to_serializable_format(value)
                for key, value in item.items()
            }
        elif isinstance(item, list):
            return [RealtimeFeaturePipeline._convert_object_to_serializable_format(value) for value in item]
        elif isinstance(item, TimestampType):
            return item.fromInternal()
        elif isinstance(item, datetime):
            return item.isoformat()
        else:
            return item

    def _node_to_value(
        self, pipeline_node: PipelineNode
    ) -> Union[
        str, int, float, bool, datetime, None, Dict[str, Any], pandas.DataFrame, PySparkDataFrame, pandas.Series
    ]:
        if pipeline_node.HasField("constant_node"):
            return constant_node_to_value(pipeline_node.constant_node)
        elif pipeline_node.HasField("feature_view_node"):
            if self._data_source_inputs is not None:
                return self._data_source_inputs[pipeline_node.feature_view_node.input_name]
            elif self.is_python_mode():
                fields_dict = {}
                # The input name of this FeatureViewNode tells us which of the udf_args
                # correspond to the Dict we should generate that the parent TransformationNode
                # expects as an input. It also expects the DataFrame to have its columns named
                # by the feature names.
                for feature in self.udf_arg_idx_map:
                    if not feature.startswith(
                        f"{udf_internal(ComputeMode.SPARK)}_{pipeline_node.feature_view_node.input_name}_{self._fv_id}"
                    ):
                        continue
                    idx = self.udf_arg_idx_map[feature]
                    fields_dict[feature_name(feature)] = self._udf_args[idx]
                return fields_dict
            elif self.is_pandas_mode():
                all_series = []
                features = []
                # The input name of this FeatureViewNode tells us which of the udf_args
                # correspond to the pandas.DataFrame we should generate that the parent TransformationNode
                # expects as an input. It also expects the DataFrame to have its columns named
                # by the feature names.
                for feature in self.udf_arg_idx_map:
                    if not feature.startswith(
                        f"{udf_internal(ComputeMode.SPARK)}_{pipeline_node.feature_view_node.input_name}_{self._fv_id}"
                    ):
                        continue
                    idx = self.udf_arg_idx_map[feature]
                    all_series.append(self._udf_args[idx])
                    features.append(feature_name(feature))
                df = pandas.concat(all_series, keys=features, axis=1)
                return df
            else:
                msg = "Transform mode {self.mode} is not yet implemented"
                raise NotImplementedError(msg)
        elif pipeline_node.HasField("request_data_source_node"):
            if self._data_source_inputs is not None:
                return self._data_source_inputs[pipeline_node.request_data_source_node.input_name]
            elif self.is_python_mode():
                request_context = pipeline_node.request_data_source_node.request_context
                field_names = [c.name for c in request_context.tecton_schema.columns]
                fields_dict = {}
                for input_col in field_names:
                    idx = self.udf_arg_idx_map[input_col]
                    fields_dict[input_col] = self._udf_args[idx]
                return fields_dict
            elif self.is_pandas_mode():
                all_series = []
                request_context = pipeline_node.request_data_source_node.request_context
                field_names = [c.name for c in request_context.tecton_schema.columns]
                for input_col in field_names:
                    idx = self.udf_arg_idx_map[input_col]
                    all_series.append(self._udf_args[idx])
                df = pandas.concat(all_series, keys=field_names, axis=1)
                return df
            else:
                msg = "Transform mode {self.mode} is not yet implemented"
                raise NotImplementedError(msg)
        elif pipeline_node.HasField("transformation_node"):
            return self._transformation_node_to_value(pipeline_node.transformation_node)
        elif pipeline_node.HasField("context_node"):
            # Used for `run_transformation` with a mock Context
            if self._data_source_inputs is not None:
                return self._data_source_inputs[pipeline_node.context_node.input_name]

            if self._events_df_timestamp_field not in self.udf_arg_idx_map:
                msg = f"Could not extract field '{self._events_df_timestamp_field}' from events data frame."
                raise Exception(msg)

            timestamp_index = self.udf_arg_idx_map[self._events_df_timestamp_field]
            request_timestamp = self._udf_args[timestamp_index]

            if self.is_python_mode():
                realtime_context = RealtimeContext(request_timestamp)
            elif self.is_pandas_mode():
                realtime_context = request_timestamp.to_frame(name="request_timestamp")
            return realtime_context
        elif pipeline_node.HasField("materialization_context_node"):
            msg = "MaterializationContext is unsupported for pandas pipelines"
            raise ValueError(msg)
        else:
            msg = "This is not yet implemented"
            raise NotImplementedError(msg)

    def _transformation_node_to_value(
        self, transformation_node: TransformationNode
    ) -> Union[Dict[str, Any], pandas.DataFrame, PySparkDataFrame]:
        """Recursively translates inputs to values and then passes them to the transformation."""
        args: List[Union[PySparkDataFrame, str, int, float, bool]] = []
        kwargs = {}
        for transformation_input in transformation_node.inputs:
            input_node = transformation_input.node
            node_value = self._node_to_value(input_node)
            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                msg = f"Unknown argument type for Input node: {transformation_input}"
                raise KeyError(msg)

        return self._apply_transformation_function(transformation_node, args, kwargs)

    def _apply_transformation_function(
        self, transformation_node: TransformationNode, args: List[Any], kwargs: Dict[str, Any]
    ) -> Union[Dict[str, Any], pandas.DataFrame, PySparkDataFrame]:
        """For the given transformation node, returns the corresponding DataFrame transformation.

        If needed, resulted function is wrapped with a function that translates mode-specific input/output types to DataFrames.
        """
        transformation = self.get_transformation_by_id(transformation_node.transformation_id)
        mode = transformation.transformation_mode
        user_function = transformation.user_function

        if (
            mode != TransformationMode.TRANSFORMATION_MODE_PANDAS
            and mode != TransformationMode.TRANSFORMATION_MODE_PYTHON
        ):
            msg = f"Unsupported transformation mode({transformation.transformation_mode}) for Realtime Feature Views."
            raise KeyError(msg)

        try:
            return user_function(*args, **kwargs)
        except TypeError as e:
            raise UDF_TYPE_ERROR(e)
        except Exception as e:
            raise UDF_ERROR(e, feature_definition_name=transformation.metadata.name)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/materialization/params__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import aws_credentials__client_pb2 as tecton__proto_dot_common_dot_aws__credentials____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.data import entity__client_pb2 as tecton__proto_dot_data_dot_entity____client__pb2
from tecton_proto.data import feature_service__client_pb2 as tecton__proto_dot_data_dot_feature__service____client__pb2
from tecton_proto.data import feature_view__client_pb2 as tecton__proto_dot_data_dot_feature__view____client__pb2
from tecton_proto.data import transformation__client_pb2 as tecton__proto_dot_data_dot_transformation____client__pb2
from tecton_proto.data import virtual_data_source__client_pb2 as tecton__proto_dot_data_dot_virtual__data__source____client__pb2
from tecton_proto.dataobs import config__client_pb2 as tecton__proto_dot_dataobs_dot_config____client__pb2
from tecton_proto.materialization import job_metadata__client_pb2 as tecton__proto_dot_materialization_dot_job__metadata____client__pb2
from tecton_proto.materialization import materialization_task__client_pb2 as tecton__proto_dot_materialization_dot_materialization__task____client__pb2
from tecton_proto.online_store_writer import config__client_pb2 as tecton__proto_dot_online__store__writer_dot_config____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1tecton_proto/materialization/params__client.proto\x12\x1ctecton_proto.materialization\x1a\x31tecton_proto/common/aws_credentials__client.proto\x1a$tecton_proto/common/id__client.proto\x1a&tecton_proto/data/entity__client.proto\x1a/tecton_proto/data/feature_service__client.proto\x1a,tecton_proto/data/feature_view__client.proto\x1a.tecton_proto/data/transformation__client.proto\x1a\x33tecton_proto/data/virtual_data_source__client.proto\x1a)tecton_proto/dataobs/config__client.proto\x1a\x37tecton_proto/materialization/job_metadata__client.proto\x1a?tecton_proto/materialization/materialization_task__client.proto\x1a\x35tecton_proto/online_store_writer/config__client.proto\"\x8c\x15\n\x19MaterializationTaskParams\x12\x41\n\x0c\x66\x65\x61ture_view\x18\x16 \x01(\x0b\x32\x1e.tecton_proto.data.FeatureViewR\x0b\x66\x65\x61tureView\x12V\n\x14virtual_data_sources\x18\x11 \x03(\x0b\x32$.tecton_proto.data.VirtualDataSourceR\x12virtualDataSources\x12K\n\x0ftransformations\x18\x17 \x03(\x0b\x32!.tecton_proto.data.TransformationR\x0ftransformations\x12\x35\n\x08\x65ntities\x18\x34 \x03(\x0b\x32\x19.tecton_proto.data.EntityR\x08\x65ntities\x12)\n\x10\x66\x65\x61ture_services\x18& \x03(\tR\x0f\x66\x65\x61tureServices\x12\x43\n\rfeature_views\x18\x38 \x03(\x0b\x32\x1e.tecton_proto.data.FeatureViewR\x0c\x66\x65\x61tureViews\x12J\n\x0f\x66\x65\x61ture_service\x18\x39 \x01(\x0b\x32!.tecton_proto.data.FeatureServiceR\x0e\x66\x65\x61tureService\x12\x36\n\x17materialization_task_id\x18# \x01(\tR\x15materializationTaskId\x12\'\n\x0fidempotence_key\x18\x12 \x01(\tR\x0eidempotenceKey\x12\x36\n\nattempt_id\x18) \x01(\x0b\x32\x17.tecton_proto.common.IdR\tattemptId\x12\x39\n\x19spark_job_execution_table\x18\x13 \x01(\tR\x16sparkJobExecutionTable\x12,\n\x12job_metadata_table\x18( \x01(\tR\x10jobMetadataTable\x12&\n\x0f\x64\x65lta_log_table\x18\x32 \x01(\tR\rdeltaLogTable\x12i\n\x17job_metadata_table_type\x18, \x01(\x0e\x32\x32.tecton_proto.materialization.JobMetadataTableTypeR\x14jobMetadataTableType\x12\x32\n\x15\x64ynamodb_table_region\x18\x14 \x01(\tR\x13\x64ynamodbTableRegion\x12}\n\x1aonline_store_writer_config\x18\x18 \x01(\x0b\x32@.tecton_proto.online_store_writer.OnlineStoreWriterConfigurationR\x17onlineStoreWriterConfig\x12^\n\x1b\x64ynamodb_cross_account_role\x18\x35 \x01(\x0b\x32\x1f.tecton_proto.common.AwsIamRoleR\x18\x64ynamodbCrossAccountRole\x12\x44\n\x1f\x64ynamodb_cross_account_role_arn\x18\r \x01(\tR\x1b\x64ynamodbCrossAccountRoleArn\x12J\n\"dynamodb_cross_account_external_id\x18\x1e \x01(\tR\x1e\x64ynamodbCrossAccountExternalId\x12\x32\n\x15\x64\x62\x66s_credentials_path\x18! \x01(\tR\x13\x64\x62\x66sCredentialsPath\x12,\n\x12offline_store_path\x18\x03 \x01(\tR\x10offlineStorePath\x12=\n\x1buse_new_consumption_metrics\x18* \x01(\x08R\x18useNewConsumptionMetrics\x12\x1b\n\tcanary_id\x18\x15 \x01(\tR\x08\x63\x61naryId\x12x\n\x19\x64\x61ta_observability_config\x18% \x01(\x0b\x32<.tecton_proto.dataobs.DataObservabilityMaterializationConfigR\x17\x64\x61taObservabilityConfig\x12U\n\x0f\x62\x61tch_task_info\x18- \x01(\x0b\x32+.tecton_proto.materialization.BatchTaskInfoH\x00R\rbatchTaskInfo\x12X\n\x10stream_task_info\x18. \x01(\x0b\x32,.tecton_proto.materialization.StreamTaskInfoH\x00R\x0estreamTaskInfo\x12X\n\x10ingest_task_info\x18/ \x01(\x0b\x32,.tecton_proto.materialization.IngestTaskInfoH\x00R\x0eingestTaskInfo\x12^\n\x12\x64\x65letion_task_info\x18\x30 \x01(\x0b\x32..tecton_proto.materialization.DeletionTaskInfoH\x00R\x10\x64\x65letionTaskInfo\x12w\n\x1b\x64\x65lta_maintenance_task_info\x18\x31 \x01(\x0b\x32\x36.tecton_proto.materialization.DeltaMaintenanceTaskInfoH\x00R\x18\x64\x65ltaMaintenanceTaskInfo\x12\x61\n\x13\x66\x65\x61ture_export_info\x18\x33 \x01(\x0b\x32/.tecton_proto.materialization.FeatureExportInfoH\x00R\x11\x66\x65\x61tureExportInfo\x12z\n\x1c\x64\x61taset_generation_task_info\x18: \x01(\x0b\x32\x37.tecton_proto.materialization.DatasetGenerationTaskInfoH\x00R\x19\x64\x61tasetGenerationTaskInfo\x12\x39\n\x17secrets_api_service_url\x18\x36 \x01(\tB\x02\x18\x01R\x14secretsApiServiceUrl\x12\x35\n\x15secret_access_api_key\x18\x37 \x01(\tB\x02\x18\x01R\x12secretAccessApiKey\x12\x30\n\x07plan_id\x18; \x01(\x0b\x32\x17.tecton_proto.common.IdR\x06planId\x12\x1e\n\x0bkms_key_arn\x18< \x01(\tR\tkmsKeyArnB\x0b\n\ttask_infoJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\tJ\x04\x08\t\x10\nJ\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0cJ\x04\x08\x0c\x10\rJ\x04\x08\x0e\x10\x0fJ\x04\x08\x0f\x10\x10J\x04\x08\x19\x10\x1aJ\x04\x08\x1a\x10\x1bJ\x04\x08\x1b\x10\x1cJ\x04\x08\x1c\x10\x1dJ\x04\x08\x1d\x10\x1eJ\x04\x08\x1f\x10 J\x04\x08 \x10!J\x04\x08\"\x10#J\x04\x08$\x10%J\x04\x08\'\x10(J\x04\x08+\x10,\"\x88\x01\n\x1fSecretMaterializationTaskParams\x12\x65\n\x15secret_service_params\x18\x01 \x01(\x0b\x32\x31.tecton_proto.materialization.SecretServiceParamsR\x13secretServiceParams\"\x7f\n\x13SecretServiceParams\x12\x35\n\x17secrets_api_service_url\x18\x01 \x01(\tR\x14secretsApiServiceUrl\x12\x31\n\x15secret_access_api_key\x18\x02 \x01(\tR\x12secretAccessApiKey\"\xb1\x02\n\rBatchTaskInfo\x12g\n\x10\x62\x61tch_parameters\x18\x01 \x01(\x0b\x32<.tecton_proto.materialization.BatchMaterializationParametersR\x0f\x62\x61tchParameters\x12\x39\n\x19\x64ynamodb_json_output_path\x18\x02 \x01(\tR\x16\x64ynamodbJsonOutputPath\x12H\n!should_dedupe_online_store_writes\x18\x03 \x01(\x08R\x1dshouldDedupeOnlineStoreWrites\x12\x32\n\x15should_avoid_coalesce\x18\x04 \x01(\x08R\x13shouldAvoidCoalesce\"\x87\x02\n\x0eStreamTaskInfo\x12j\n\x11stream_parameters\x18\x01 \x01(\x0b\x32=.tecton_proto.materialization.StreamMaterializationParametersR\x10streamParameters\x12:\n\x19streaming_checkpoint_path\x18\x02 \x01(\tR\x17streamingCheckpointPath\x12M\n#streaming_trigger_interval_override\x18\x03 \x01(\tR streamingTriggerIntervalOverride\"|\n\x0eIngestTaskInfo\x12j\n\x11ingest_parameters\x18\x01 \x01(\x0b\x32=.tecton_proto.materialization.IngestMaterializationParametersR\x10ingestParameters\"u\n\x10\x44\x65letionTaskInfo\x12\x61\n\x13\x64\x65letion_parameters\x18\x01 \x01(\x0b\x32\x30.tecton_proto.materialization.DeletionParametersR\x12\x64\x65letionParameters\"\x96\x01\n\x18\x44\x65ltaMaintenanceTaskInfo\x12z\n\x1c\x64\x65lta_maintenance_parameters\x18\x01 \x01(\x0b\x32\x38.tecton_proto.materialization.DeltaMaintenanceParametersR\x1a\x64\x65ltaMaintenanceParameters\"\x86\x01\n\x11\x46\x65\x61tureExportInfo\x12q\n\x19\x66\x65\x61ture_export_parameters\x18\x01 \x01(\x0b\x32\x35.tecton_proto.materialization.FeatureExportParametersR\x17\x66\x65\x61tureExportParameters\"\x9a\x01\n\x19\x44\x61tasetGenerationTaskInfo\x12}\n\x1d\x64\x61taset_generation_parameters\x18\x01 \x01(\x0b\x32\x39.tecton_proto.materialization.DatasetGenerationParametersR\x1b\x64\x61tasetGenerationParametersB\x1e\n\x1a\x63om.tecton.materializationP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.materialization.params__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\032com.tecton.materializationP\001'
  _MATERIALIZATIONTASKPARAMS.fields_by_name['secrets_api_service_url']._options = None
  _MATERIALIZATIONTASKPARAMS.fields_by_name['secrets_api_service_url']._serialized_options = b'\030\001'
  _MATERIALIZATIONTASKPARAMS.fields_by_name['secret_access_api_key']._options = None
  _MATERIALIZATIONTASKPARAMS.fields_by_name['secret_access_api_key']._serialized_options = b'\030\001'
  _MATERIALIZATIONTASKPARAMS._serialized_start=629
  _MATERIALIZATIONTASKPARAMS._serialized_end=3329
  _SECRETMATERIALIZATIONTASKPARAMS._serialized_start=3332
  _SECRETMATERIALIZATIONTASKPARAMS._serialized_end=3468
  _SECRETSERVICEPARAMS._serialized_start=3470
  _SECRETSERVICEPARAMS._serialized_end=3597
  _BATCHTASKINFO._serialized_start=3600
  _BATCHTASKINFO._serialized_end=3905
  _STREAMTASKINFO._serialized_start=3908
  _STREAMTASKINFO._serialized_end=4171
  _INGESTTASKINFO._serialized_start=4173
  _INGESTTASKINFO._serialized_end=4297
  _DELETIONTASKINFO._serialized_start=4299
  _DELETIONTASKINFO._serialized_end=4416
  _DELTAMAINTENANCETASKINFO._serialized_start=4419
  _DELTAMAINTENANCETASKINFO._serialized_end=4569
  _FEATUREEXPORTINFO._serialized_start=4572
  _FEATUREEXPORTINFO._serialized_end=4706
  _DATASETGENERATIONTASKINFO._serialized_start=4709
  _DATASETGENERATIONTASKINFO._serialized_end=4863
# @@protoc_insertion_point(module_scope)



import abc
from typing import List

from pypika.enums import SqlTypes

from tecton_proto.common import data_type__client_pb2 as data_type_pb2


class DataType(abc.ABC):
    """
    Base DataType class. A python utility for working with `tecton_proto.common.DataType` protos.
    """

    @property
    @abc.abstractmethod
    def proto(self) -> data_type_pb2.DataType:
        pass

    def __hash__(self) -> int:
        return hash(self.proto.SerializeToString(deterministic=True))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DataType):
            return False
        return self.proto == other.proto

    def __repr__(self):
        # Require __repr__ implementation. Used in error messages.
        raise NotImplementedError

    # TODO: if we stop throwing NotImplementedError, this could be made a real attribute
    @property
    def sql_type(self) -> SqlTypes:
        # Used for safe typecasting in sql-based compute platforms
        raise NotImplementedError


class UnknownType(DataType):
    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(type=data_type_pb2.DATA_TYPE_UNKNOWN)

    def __repr__(self) -> str:
        return "Unknown"

    @property
    def sql_type(self) -> SqlTypes:
        msg = "pypika doesn't support `UNKNOWN` type"
        raise NotImplementedError(msg)


class Int32Type(DataType):
    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(type=data_type_pb2.DATA_TYPE_INT32)

    def __repr__(self) -> str:
        return "Int32"

    @property
    def sql_type(self) -> SqlTypes:
        return SqlTypes.INTEGER


class Int64Type(DataType):
    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(type=data_type_pb2.DATA_TYPE_INT64)

    def __repr__(self) -> str:
        return "Int64"

    @property
    def sql_type(self) -> SqlTypes:
        return SqlTypes.INTEGER


class Float32Type(DataType):
    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(type=data_type_pb2.DATA_TYPE_FLOAT32)

    def __repr__(self) -> str:
        return "Float32"

    @property
    def sql_type(self) -> SqlTypes:
        return SqlTypes.FLOAT


class Float64Type(DataType):
    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(type=data_type_pb2.DATA_TYPE_FLOAT64)

    def __repr__(self) -> str:
        return "Float64"

    @property
    def sql_type(self) -> SqlTypes:
        return SqlTypes.FLOAT


class StringType(DataType):
    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(type=data_type_pb2.DATA_TYPE_STRING)

    def __repr__(self) -> str:
        return "String"

    @property
    def sql_type(self) -> SqlTypes:
        return SqlTypes.VARCHAR


class BoolType(DataType):
    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(type=data_type_pb2.DATA_TYPE_BOOL)

    def __repr__(self) -> str:
        return "Bool"

    @property
    def sql_type(self) -> SqlTypes:
        return SqlTypes.BOOLEAN


class TimestampType(DataType):
    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(type=data_type_pb2.DATA_TYPE_TIMESTAMP)

    def __repr__(self) -> str:
        return "Timestamp"

    @property
    def sql_type(self) -> SqlTypes:
        return SqlTypes.TIMESTAMP


class ArrayType(DataType):
    def __init__(self, element_type: DataType) -> None:
        self._element_type = element_type

    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(type=data_type_pb2.DATA_TYPE_ARRAY, array_element_type=self.element_type.proto)

    @property
    def element_type(self) -> DataType:
        return self._element_type

    def __repr__(self) -> str:
        return f"Array({self._element_type})"

    # TODO: have this work for SQL types
    @property
    def sql_type(self) -> SqlTypes:
        # Used for safe typecasting in sql-based compute platforms
        raise NotImplementedError


# Note StructField does not inherit from DataType. This is because it is not directly convertable to a data type proto.
class StructField:
    def __init__(self, name: str, data_type: DataType) -> None:
        self._name = name
        self._data_type = data_type

    @property
    def proto(self) -> data_type_pb2.StructField:
        return data_type_pb2.StructField(name=self._name, data_type=self._data_type.proto)

    @property
    def name(self) -> str:
        return self._name

    @property
    def data_type(self) -> DataType:
        return self._data_type

    def __repr__(self) -> str:
        return f"Field({self._name}, {self._data_type})"


class StructType(DataType):
    def __init__(self, fields: List[StructField]) -> None:
        self._fields = fields

    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(
            type=data_type_pb2.DATA_TYPE_STRUCT, struct_fields=[field.proto for field in self.fields]
        )

    @property
    def fields(self) -> List[StructField]:
        return self._fields

    def __repr__(self) -> str:
        fields_string = ", ".join(str(field) for field in self._fields)
        return f"Struct({fields_string})"

    # TODO: have this work for SQL types
    @property
    def sql_type(self) -> SqlTypes:
        # Used for safe typecasting in sql-based compute platforms
        raise NotImplementedError


class MapType(DataType):
    def __init__(self, key_type: DataType, value_type: DataType) -> None:
        self._key_type = key_type
        self._value_type = value_type

    @property
    def proto(self) -> data_type_pb2.DataType:
        return data_type_pb2.DataType(
            type=data_type_pb2.DATA_TYPE_MAP, map_key_type=self._key_type.proto, map_value_type=self._value_type.proto
        )

    @property
    def key_type(self) -> DataType:
        return self._key_type

    @property
    def value_type(self) -> DataType:
        return self._value_type


def data_type_from_proto(proto: data_type_pb2.DataType) -> DataType:
    """
    Factory method to creata a DataType python class from a `tecton_proto.common.DataType` proto.
    """
    assert proto

    if proto.type == data_type_pb2.DATA_TYPE_INT32:
        return Int32Type()
    elif proto.type == data_type_pb2.DATA_TYPE_INT64:
        return Int64Type()
    elif proto.type == data_type_pb2.DATA_TYPE_FLOAT32:
        return Float32Type()
    elif proto.type == data_type_pb2.DATA_TYPE_FLOAT64:
        return Float64Type()
    elif proto.type == data_type_pb2.DATA_TYPE_STRING:
        return StringType()
    elif proto.type == data_type_pb2.DATA_TYPE_BOOL:
        return BoolType()
    elif proto.type == data_type_pb2.DATA_TYPE_TIMESTAMP:
        return TimestampType()
    elif proto.type == data_type_pb2.DATA_TYPE_ARRAY:
        assert proto.array_element_type
        element_type = data_type_from_proto(proto.array_element_type)
        return ArrayType(element_type)
    elif proto.type == data_type_pb2.DATA_TYPE_STRUCT:
        fields = [StructField(field.name, data_type_from_proto(field.data_type)) for field in proto.struct_fields]
        return StructType(fields)
    elif proto.type == data_type_pb2.DATA_TYPE_MAP:
        return MapType(data_type_from_proto(proto.map_key_type), data_type_from_proto(proto.map_value_type))
    elif proto.type == data_type_pb2.DATA_TYPE_UNKNOWN:
        return UnknownType()
    else:
        msg = f"Unexpected data type {proto}"
        raise ValueError(msg)



from typing import Union

import pandas
import pyarrow
import pyarrow.dataset


# These are kwargs for the pyarrow.dataset.ParquetFileFormat.make_write_options function. They are "unwrapped" like this
# because some other APIs (e.g. pyarrow.parquet.write_to_dataset, which is in turn called by
# pandas.DataFrame.to_parquet) also accept write options in kwargs format but not FileWriteOptions format.
PARQUET_WRITE_OPTIONS_KWARGS = {
    "version": "2.4",
    "data_page_version": "1.0",
    "compression": "snappy",
}

PARQUET_WRITE_OPTIONS = pyarrow.dataset.ParquetFileFormat().make_write_options(**PARQUET_WRITE_OPTIONS_KWARGS)


def arrow_to_pandas_dataframe(table: Union[pyarrow.Table, pyarrow.RecordBatchReader]) -> pandas.DataFrame:
    if isinstance(table, pyarrow.RecordBatchReader):
        to_pandas = table.read_pandas
    else:
        to_pandas = table.to_pandas

    df = to_pandas(
        types_mapper={
            # until we upgrade pandas (and move to arrow-based type system)
            # we need to handle int columns that might contain nulls
            pyarrow.int32(): pandas.Int32Dtype(),
            pyarrow.int64(): pandas.Int64Dtype(),
        }.get
    )
    for col in df:
        if pandas.notnull(df[col]).all():
            continue

        # If column contains nulls (ie, numpy.nan or pandas.NA)
        # convert this column type to object to replace these nulls with Python None.
        # This is needed to match transform server, which converts all nulls to Nones.
        df[col] = df[col].astype("object")
        df[col] = df[col].where(pandas.notnull(df[col]), None)

    return df



import datetime
import logging
from typing import Dict
from typing import Optional
from typing import Set

import attrs
from typeguard import typechecked

from tecton._internals import sdk_decorators
from tecton_core import id_helper
from tecton_core import specs
from tecton_proto.args import basic_info__client_pb2 as basic_info_pb2
from tecton_proto.args import fco_args__client_pb2 as fco_args_pb2
from tecton_proto.args import repo_metadata__client_pb2 as repo_metadata_pb2
from tecton_proto.common import id__client_pb2 as id_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


logger = logging.getLogger(__name__)

_LOCAL_TECTON_OBJECTS: Set["BaseTectonObject"] = set()


@attrs.frozen
class TectonObjectInfo:
    """A public SDK dataclass containing common metadata used for all Tecton Objects."""

    id: str
    name: str
    description: Optional[str]
    tags: Dict[str, str]
    owner: Optional[str]
    workspace: Optional[str]
    created_at: Optional[datetime.datetime] = attrs.field(repr=False)
    defined_in: Optional[str] = attrs.field(repr=False)
    _is_local_object: bool = attrs.field(repr=False)

    @classmethod
    @typechecked
    def from_args_proto(cls, basic_info: basic_info_pb2.BasicInfo, id: id_pb2.Id) -> "TectonObjectInfo":
        return cls(
            id=id_helper.IdHelper.to_string(id),
            name=basic_info.name,
            description=basic_info.description if basic_info.HasField("description") else None,
            tags=dict(basic_info.tags),
            owner=basic_info.owner if basic_info.HasField("owner") else None,
            created_at=None,  # created_at is only filled for remote (i.e. applied) Tecton objects.
            workspace=None,  # workspace is only filled for remote (i.e. applied) Tecton objects.
            defined_in=None,  # defined_in is only filled for remote (i.e. applied) Tecton objects.
            is_local_object=True,
        )

    @classmethod
    @typechecked
    def from_spec(cls, spec: specs.TectonObjectSpec) -> "TectonObjectInfo":
        return cls(
            id=spec.id,
            name=spec.name,
            description=spec.metadata.description,
            tags=spec.metadata.tags,
            owner=spec.metadata.owner,
            created_at=spec.metadata.created_at,
            workspace=spec.workspace,
            defined_in=spec.metadata.defined_in,
            is_local_object=spec.is_local_object,
        )

    @property
    def _id_proto(self) -> id_pb2.Id:
        return id_helper.IdHelper.from_string(self.id)


@attrs.define(eq=False)
class BaseTectonObject:
    """The base class for all Tecton Objects, e.g. Entities, Data Sources, and Feature Views.

    Attributes:
        info: A dataclass containing basic info about this Tecton Object.
    """

    info: TectonObjectInfo = attrs.field(on_setattr=attrs.setters.frozen)

    # Metadata about where this object was defined in the repo, e.g. the filename and line number. Only set if this
    # object was defined locally.
    _source_info: Optional[repo_metadata_pb2.SourceInfo] = attrs.field(repr=False, on_setattr=attrs.setters.frozen)

    def _validate(self, indentation_level: int = 0) -> None:
        """Internal validation API."""
        raise NotImplementedError

    @sdk_decorators.sdk_public_method  # Not actually a "public" method but is logged for analytics.
    def _run_automatic_validation(self) -> None:
        self._validate()

    @sdk_decorators.sdk_public_method
    def validate(self) -> None:
        """Deprecated"""
        # TODO(TEC-19865): revisit what to do with this function as part of schema validation
        logger.warning(
            "The validate() method is deprecated and will be removed in a future version. As of Tecton "
            "version 1.0 objects are validated upon object creation, so validate() is unnecessary."
        )

    def _build_args(self) -> fco_args_pb2.FcoArgs:
        """Returns a copy of the args as a FcoArgs proto for plan/apply logic."""
        raise NotImplementedError

    def _build_fco_validation_args(self) -> validator_pb2.FcoValidationArgs:
        """Returns a copy of the args as a FcoValidationArgs proto for validation logic."""
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Returns the name of the Tecton object."""
        return self.info.name

    @property
    def id(self) -> str:
        """Returns the unique id of the Tecton object."""
        return self.info.id

    @property
    def description(self) -> Optional[str]:
        """Returns the description of the Tecton object."""
        return self.info.description

    @property
    def tags(self) -> Dict[str, str]:
        """Returns the tags of the Tecton object."""
        return self.info.tags

    @property
    def owner(self) -> Optional[str]:
        """Returns the owner of the Tecton object."""
        return self.info.owner

    @property
    def workspace(self) -> Optional[str]:
        """Returns the workspace that this Tecton object belongs to. `None` for locally defined objects."""
        return self.info.workspace

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        """Returns the time that this Tecton object was created or last updated. `None` for locally defined objects."""
        return self.info.created_at

    @property
    def defined_in(self) -> Optional[str]:
        """The repo filename where this object was declared. `None` for locally defined objects."""
        return self.info.defined_in

    @property
    def _is_local_object(self) -> bool:
        """Returns True if the object was defined locally, i.e. was not applied and fetched from the Tecton backend."""
        return self.info._is_local_object

    @property
    def _id_proto(self) -> id_pb2.Id:
        """Returns the proto version of the Tecton object id."""
        return self.info._id_proto


def _register_local_object(obj: BaseTectonObject) -> None:
    """Register the tecton object to the set of global tecton objects.

    The global set is used to collect objects during `tecton apply`.
    """
    _LOCAL_TECTON_OBJECTS.add(obj)



from typing import Dict
from typing import List
from typing import Optional

from tecton._internals.tecton_pydantic import StrictModel
from tecton.types import Field


_CUSTOM_MODELS: List["ModelConfig"] = []


class ModelConfig(StrictModel):
    """A Custom Model in Tecton

    Example of an Model declaration:

    .. code-block:: python

        from tecton import ModelConfig

        model_config = ModelConfig(
            name=”mymodel_text_embedder”,
            model_type="pytorch",
            model_file="model_file.py",
            environments=["custom-ml-env-01"],
            artifact_files=["model.safetensors", "tokenizer_config.json" …],
            input_schema=[Field(”my_text_col”, String), …],
            output_schema=Field(”text_embedding”)
        )
    """

    name: str
    model_type: str
    description: Optional[str]
    tags: Optional[Dict[str, str]]
    model_file: str
    artifact_files: List[str]
    environments: List[str]
    input_schema: List[Field]
    output_schema: Field

    def __init__(
        self,
        *,
        name: str,
        model_type: str,
        model_file: str,
        artifact_files: List[str],
        environments: List[str],
        input_schema: List[Field],
        output_schema: Field,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        """Declare a new Model.

        :param name: Unique name of model
        :param description: Description of Model
        :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
        :param model_type: Type of Model (pytorch)
        :param model_file: Relative Path of File containing model.
        :param artifact_files: Relative Path of other files needed to load and run model.
        :param environments: All the environments allowed for this custom model to run in.
        :param input_schema: Input Schema for model.
        :param output_schema: Output Schema of model.

        :raises TectonValidationError: if the input non-parameters are invalid.
        """
        super().__init__(
            name=name,
            model_type=model_type,
            model_file=model_file,
            artifact_files=artifact_files,
            environments=environments,
            input_schema=input_schema,
            output_schema=output_schema,
            description=description,
            tags=tags,
        )
        _CUSTOM_MODELS.append(self)



import datetime
import re
from typing import Optional
from typing import TypeVar
from typing import Union

import pandas
import pendulum
import pytimeparse
from google.protobuf import duration_pb2
from google.protobuf import timestamp_pb2
from pypika.terms import Term
from typeguard import typechecked

from tecton_core.errors import TectonValidationError
from tecton_proto.data import feature_store__client_pb2 as feature_store_pb2


# 10,000 years which is less than but still close to the max duration proto amount, and a digestable number of years to
# a user
_MAX_PROTO_DURATION_SECONDS = 10_000 * 365 * 24 * 60 * 60

_TIME_RANGE_REGEX = r"^(?:(\d+)d)?(?:(\d+)h)?(?:(\d+)m)?(?:(\d+)s)?$"

TERM_OR_INT = TypeVar("TERM_OR_INT", int, Term)


def timedelta_to_duration(td: datetime.timedelta) -> pendulum.Duration:
    return pendulum.duration(days=td.days, seconds=td.seconds, microseconds=td.microseconds)


def proto_to_duration(proto_duration: duration_pb2.Duration) -> pendulum.Duration:
    return timedelta_to_duration(proto_duration.ToTimedelta())


def timedelta_to_proto(td: Optional[datetime.timedelta]) -> Optional[duration_pb2.Duration]:
    if td is None:
        return None

    if td.total_seconds() > _MAX_PROTO_DURATION_SECONDS:
        msg = "'Timedelta value must be lower than max value of 10000 years'"
        raise TectonValidationError(msg)

    proto = duration_pb2.Duration()
    proto.FromTimedelta(td)
    return proto


def to_positive_proto(proto_duration: duration_pb2.Duration) -> duration_pb2.Duration:
    return duration_pb2.Duration(seconds=abs(duration_pb2.Duration.seconds), nanos=abs(duration_pb2.Duration.nanos))


def datetime_to_proto(dt: Optional[datetime.datetime]) -> Optional[timestamp_pb2.Timestamp]:
    if dt is None:
        return None
    proto = timestamp_pb2.Timestamp()
    proto.FromDatetime(dt)
    return proto


@typechecked
def to_human_readable_str(duration: Union[duration_pb2.Duration, datetime.timedelta]) -> str:
    """
    Replicate the backend logic that converts durations to feature strings, e.g. `column_sum_1h30m_30m`.

    Backend logic: https://github.com/tecton-ai/tecton/blob/e3c04e28066c21fa4ed9b459ece458c084766b6e/java/com/tecton/common/utils/TimeUtils.kt#L80
    """

    if isinstance(duration, duration_pb2.Duration):
        duration = proto_to_duration(duration)
    elif isinstance(duration, datetime.timedelta):
        duration = timedelta_to_duration(duration)
    else:
        msg = f"Invalid input duration type: {type(duration)}"
        raise TypeError(msg)

    duration_str_list = []
    if duration.days > 0:
        duration_str_list.append(f"{duration.days}d")
    if duration.hours > 0:
        duration_str_list.append(f"{duration.hours}h")
    if duration.minutes > 0:
        duration_str_list.append(f"{duration.minutes}m")
    if duration.remaining_seconds > 0:
        duration_str_list.append(f"{duration.remaining_seconds}s")

    return "0s" if len(duration_str_list) == 0 else "".join(duration_str_list)


@typechecked
def parse_to_timedelta(time_string: str) -> datetime.timedelta:
    """Parse a string with format '{days}d{hours}h{minutes}m{seconds}s' into a timedelta object.

    Each time component is optional, but at least one value must be provided, e.g. "" is invalid but "0s" is valid.
    """
    if time_string == "":
        msg = "Cannot use an empty string for a time range. Use '0d', '0s', etc. instead."
        raise ValueError(msg)

    match = re.match(_TIME_RANGE_REGEX, time_string)
    if not match:
        msg = f"The time range string '{time_string}' is not in the correct format. Should be of the form '[days]d[hours]h[minutes]m[seconds]s', e.g. `1d12h`."
        raise ValueError(msg)

    # Extract time components from the matched groups, substituting 0 where no component is found
    days, hours, minutes, seconds = [int(g) if g else 0 for g in match.groups()]

    # Create a timedelta object with the extracted time components
    return datetime.timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)


def assert_is_round_seconds(d: pendulum.Duration) -> pendulum.Duration:
    if d % pendulum.Duration(seconds=1):
        msg = f"{d.in_words()} is not a round number of seconds"
        raise ValueError(msg)
    return d


def strict_pytimeparse(time_str: str) -> Union[int, float]:
    parsed = pytimeparse.parse(time_str)
    if parsed is None:
        msg = f'Could not parse time string "{time_str}"'
        raise TectonValidationError(msg)
    else:
        return parsed


def nanos_to_seconds(nanos: int) -> float:
    """
    :param nanos: Nanoseconds
    :return: Converts nanoseconds to seconds
    """
    return nanos / float(1e9)


def seconds_to_nanos(seconds: float) -> int:
    """
    :param seconds: Seconds
    :return: Converts seconds to nanoseconds
    """
    return int(seconds * 1e9)


def convert_duration_to_seconds(duration: int, version: int) -> int:
    """Convert a duration to seconds according to the version.

    If V0, the duration is already in seconds; if V1, the duration must be converted from nanos to seconds.
    """
    if version == feature_store_pb2.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return duration
    else:
        return int(nanos_to_seconds(duration))


def convert_timedelta_for_version(duration: datetime.timedelta, version: int) -> int:
    """
    Convert pendulum duration according to version
    VO -> Return Seconds
    V1 -> Return Nanoseconds
    :param duration: Pendulum Duration
    :param version: Feature Store Format Version
    :return:
    """
    assert duration.microseconds == 0
    interval = duration.total_seconds()
    if version == feature_store_pb2.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return int(interval)
    else:
        return seconds_to_nanos(interval)


def convert_epoch_to_datetime(epoch: int, version: int) -> pendulum.datetime:
    """
    Converts an epoch to a datetime.
    """
    if version == feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return pendulum.from_timestamp(epoch)
    else:
        return pendulum.from_timestamp(epoch / 1e9)


def convert_timestamp_for_version(timestamp: datetime.datetime, version: int) -> int:
    if version == feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return int(timestamp.timestamp())
    else:
        return seconds_to_nanos(timestamp.timestamp())


def convert_proto_duration_for_version(duration: duration_pb2.Duration, version: int) -> int:
    return convert_timedelta_for_version(duration.ToTimedelta(), version)


def align_time_downwards(time: datetime.datetime, alignment: datetime.timedelta) -> datetime.datetime:
    excess_seconds = time.timestamp() % alignment.total_seconds()
    return datetime.datetime.utcfromtimestamp(time.timestamp() - excess_seconds)


def align_epoch_downwards(int_timestamp_col: int, window_size: int) -> int:
    return int_timestamp_col - (int_timestamp_col % window_size)


def align_time_upwards(time: datetime.datetime, alignment: datetime.timedelta) -> datetime.datetime:
    excess_seconds = time.timestamp() % alignment.total_seconds()
    offset = 0 if excess_seconds == 0 else alignment.total_seconds() - excess_seconds
    return datetime.datetime.utcfromtimestamp(time.timestamp() + offset)


def get_timezone_aware_datetime(time: Optional[datetime.datetime]) -> Optional[datetime.datetime]:
    """
    Check and get timezone aware datetime
    :param time: Time to convert to timezone aware
    """
    if time is None:
        return None
    if time.tzinfo is None:
        return time.replace(tzinfo=pendulum.timezone("UTC"))
    else:
        return time


def convert_pandas_df_for_snowflake_upload(pandas_df: pandas.DataFrame) -> None:
    """
    Converts a pandas dataframe's timestamp columns to ensure expected behavior when uploading to Snowflake.

    Snowflake's `write_pandas` logic only writes TIMESTAMP_NTZ timestamps correctly if we
    pretend the input timestamps are UTC. Providing another time zone (e.g. 11PM PST) will convert
    this to UTC time (6AM UTC), and then drop the timezone info + load as a TIMESTAMP_NTZ column (6AM).
    """
    # TODO(TEC-16678): Revert this once a fix is upstreamed to Snowflake
    for col in pandas_df.columns:
        if pandas.core.dtypes.common.is_datetime64_dtype(pandas_df[col]):
            pandas_df[col] = pandas_df[col].dt.tz_localize(None)
            pandas_df[col] = pandas_df[col].dt.tz_localize(pendulum.timezone("UTC"))


def convert_to_effective_timestamp(
    timestamp_col: TERM_OR_INT, batch_schedule_seconds: int, data_delay_seconds: int
) -> TERM_OR_INT:
    """
    Converts a timestamp col term or integer and converts it into an effective timestamp by aligning on the interval
    and adding the interval and data delay.

    The effective timestamp represents the time when the feature is available in the online store to be used.
    """
    return timestamp_col - (timestamp_col % batch_schedule_seconds) + batch_schedule_seconds + data_delay_seconds


def get_nearest_anchor_time(
    timestamp: pendulum.DateTime,
    max_source_data_delay: pendulum.Duration,
    batch_materialization_schedule: pendulum.Duration,
    min_scheduling_interval: pendulum.Duration,
) -> pendulum.DateTime:
    """
    Returns the nearest anchor time (start of tile), such that the feature values as of this time would have the
    effective timestamp `timestamp`
    """
    timestamp = timestamp - max_source_data_delay - batch_materialization_schedule

    if min_scheduling_interval.total_seconds() > 0:
        timestamp = align_time_downwards(timestamp, min_scheduling_interval)

    return pendulum.instance(timestamp)



from typing import Callable
from typing import List
from typing import Optional
from typing import Union

import attrs
import pyspark.sql.functions as F
from pyspark import SparkContext
from pyspark.sql import Column
from pyspark.sql import WindowSpec
from pyspark.sql import functions
from pyspark.sql.functions import expr

from tecton_core.aggregation_utils import get_materialization_aggregation_column_prefixes
from tecton_core.data_types import DataType
from tecton_core.data_types import Int64Type
from tecton_core.data_types import StringType
from tecton_proto.common import aggregation_function__client_pb2 as afpb
from tecton_spark.jars.class_loader import get_or_create_udf_jar_class_loader


# WARNING: If you're changing this class there's a good chance you need to change
# AggregationPlans.kt. Please look over that file carefully.


@attrs.define(auto_attribs=True)
class AggregationPlan(object):
    """
    An AggregationPlan contains all the methods required to compute feature values for a specific Tecton aggregation.

    Partial aggregates are returned as a list of pyspark columns. A single column is insufficient since certain
    aggregations require multiple intermediate columns. For example, the mean aggregation needs sum and count columns.

    Full aggregates use the partial aggregate columns as inputs and are returned as a single pyspark column.

    The order of the columns must be the same in:
    * the return list in partial_aggregation_transform
    * the return list in continuous_partial_aggregation_transform
    * the arguments list in full_aggregation_transform
    * materialized_column_prefixes

    Similarly, if there exists a partial_aggregation_preprocessor, the order of the columns must be the same in:
    * the output columns of partial_aggregation_preprocessor
    * the input columns to partial_aggregation_transform
    * intermediate_column_prefixes

    Attributes:
        partial_aggregation_transform: A method that maps a list of input column names to a list of output pyspark columns containing the partial aggregates.
        continuous_partial_aggregation_transform: Maps a list of input column names to a list of output pyspark columns.
        full_aggregation_transform: A method that maps a list of input partial aggregate columns and a WindowSpec to an output pyspark column containing the full aggregates.
        materialized_column_prefixes: The list of prefixes that should be applied to the pyspark columns produced by `partial_aggregation_transform`.
            containing the partial aggregates for continuous mode.
        partial_aggregation_preprocessor: Creates intermediate columns to be used by the partial_aggregation_transform.
            This should only be necessary when the partial aggregation computation needs to be split across pyspark and
            a UDAF, as is the case for the approx count distinct aggregation. If it exists, it is applied in both
            continuous and non-continuous mode.
        intermediate_column_prefixes: The list of prefixes that should be applied to the pyspark columns produced by `partial_aggregation_preprocessor`.
    """

    partial_aggregation_transform: Callable[[List[str]], List[Column]]
    continuous_partial_aggregation_transform: Callable[[List[str]], List[Column]]
    full_aggregation_transform: Callable[[List[str], WindowSpec], Column]
    materialized_column_prefixes: List[str]
    partial_aggregation_preprocessor: Optional[Callable[[str], List[Column]]] = None
    intermediate_column_prefixes: Optional[List[str]] = None

    def materialized_column_names(self, input_column_name: str) -> List[str]:
        return [f"{prefix}_{input_column_name}" for prefix in self.materialized_column_prefixes]

    def intermediate_column_names(self, input_column_name: str) -> List[str]:
        return [f"{prefix}_{input_column_name}" for prefix in self.intermediate_column_prefixes]


def get_aggregation_plan(
    aggregation_function: afpb.AggregationFunction,
    function_params: afpb.AggregationFunctionParams,
    is_continuous: bool,
    time_key: str,
    col_datatype: DataType = None,
) -> AggregationPlan:
    plan = AGGREGATION_PLANS.get(aggregation_function, None)
    if plan is None:
        msg = f"Unsupported aggregation function {aggregation_function}"
        raise ValueError(msg)

    if callable(plan):
        return plan(time_key, function_params, is_continuous, col_datatype)
    else:
        return plan


def _simple_partial_aggregation_transform(spark_transform):
    return lambda cols: [spark_transform(cols[0])]


def _simple_continuous_partial_aggregation_transform():
    return lambda cols: [F.col(cols[0])]


def _simple_full_aggregation_transform(spark_transform):
    return lambda cols, window: spark_transform(cols[0]).over(window)


def _simple_aggregation_plan(
    aggregation_function: afpb.AggregationFunction, spark_transform: Callable[[Union[Column, str]], Column]
) -> AggregationPlan:
    return AggregationPlan(
        partial_aggregation_transform=_simple_partial_aggregation_transform(spark_transform),
        continuous_partial_aggregation_transform=_simple_continuous_partial_aggregation_transform(),
        full_aggregation_transform=_simple_full_aggregation_transform(spark_transform),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(aggregation_function),
    )


def ApproxPercentilePartialAgg(col: str, precision: int) -> List[Column]:
    sc = SparkContext._active_spark_context
    udf_name = f"tecton_approx_percentile_partial_{precision}"
    get_or_create_udf_jar_class_loader().load_class(
        sc, "com.tecton.udfs.spark3.ApproxPercentilePartialRegister"
    ).newInstance().register(udf_name, precision)
    columns = expr(f"{udf_name}({col})")
    return [columns.processedMeans, columns.processedWeights]


def _make_approx_percentile_partial_aggregation(precision: int) -> Callable[[List[str]], List[Column]]:
    def _approx_percentile_partial_aggregation(cols: List[str]) -> List[Column]:
        return ApproxPercentilePartialAgg(cols[0], precision)

    return _approx_percentile_partial_aggregation


def ApproxPercentileFullAgg(
    processed_means: str,
    processed_weights: str,
    percentile: float,
    precision: int,
) -> Column:
    sc = SparkContext._active_spark_context
    percentile_str = str(percentile).replace(".", "_")
    udf_name = f"tecton_approx_percentile_full_{percentile_str}"
    get_or_create_udf_jar_class_loader().load_class(
        sc, "com.tecton.udfs.spark3.ApproxPercentileFullRegister"
    ).newInstance().register(udf_name, precision, percentile)
    return expr(f"{udf_name}({processed_means}, {processed_weights})")


def _make_approx_percentile_full_aggregation(percentile: float, precision: int) -> Callable[[List[str]], WindowSpec]:
    def _approx_percentile_full_aggregation(cols: List[str], window: WindowSpec) -> Column:
        processed_means, processed_weights = cols
        return (
            ApproxPercentileFullAgg(processed_means, processed_weights, percentile, precision).over(window).percentile
        )

    return _approx_percentile_full_aggregation


def _make_approx_percentile_continuous_partial() -> Callable[[List[str]], List[Column]]:
    def _approx_percentile_continuous_partial(cols: List[str]) -> List[Column]:
        input_col = F.col(cols[0])

        # The partial aggregate tiles for approx_percentile do not allow nulls, so we replace nulls with NaNs, which will be filtered out of the eventual result.
        mean_col = F.when(input_col.isNull(), F.lit(float("nan"))).otherwise(input_col)
        return [F.array(mean_col), F.array(F.lit(1.0))]

    return _approx_percentile_continuous_partial


def ApproxCountDistinctFullAgg(indices: str, registers: str, precision: int) -> Column:
    sc = SparkContext._active_spark_context
    udf_name = "tecton_approx_count_distinct_full"
    get_or_create_udf_jar_class_loader().load_class(
        sc, "com.tecton.udfs.spark3.ApproxCountDistinctFullRegister"
    ).newInstance().register(udf_name, precision)
    return expr(f"{udf_name}({indices}, {registers})")


def _make_approx_count_distinct_full_aggregation(precision: int) -> Callable:
    def _approx_count_distinct_full_aggregation(cols: List[str], window: WindowSpec) -> Column:
        indices, registers = cols[0], cols[1]
        return ApproxCountDistinctFullAgg(indices, registers, precision).over(window).count

    return _approx_count_distinct_full_aggregation


def ApproxCountDistinctPartialAgg(index: str, register: str) -> List[Column]:
    sc = SparkContext._active_spark_context
    udf_name = "tecton_approx_count_distinct_partial"
    get_or_create_udf_jar_class_loader().load_class(
        sc, "com.tecton.udfs.spark3.ApproxCountDistinctPartialRegister"
    ).newInstance().register(udf_name)
    columns = expr(f"{udf_name}({index}, {register})")
    return [columns.indices, columns.registers]


def _make_approx_count_distinct_partial_aggregation() -> Callable[[List[str]], Column]:
    def _approx_count_distinct_partial_aggregation(cols: List[str]) -> List[Column]:
        index, register = cols[0], cols[1]
        return ApproxCountDistinctPartialAgg(index, register)

    return _approx_count_distinct_partial_aggregation


def _make_approx_count_distinct_partial_aggregation_helper(precision: int) -> Callable[[str], List[Column]]:
    def _add_index_and_register_column(input_column: str) -> List[Column]:
        """Computes the index and register values and returns them as bigint columns.

        The input column is first hashed with the SHA256 hash function and truncated to 64 bits. Then:
        * The index is the first 'precision' bits of the hash.
        * The register value is 1 + (number of leading zeros in the final 64 - 'precision' bits of the hash).

        For example, if the the input column has an entry "foo" and the precision is 8:
        * The SHA256 hash is '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae'.
        * The first 64 bits are '2c26b46b68ffc68f', or '0010110000100110101101000110101101101000111111111100011010001111' in binary.
        * The first 8 bits are '00101100', or 44 in decimal. This is the index.
        * The last 56 bits are '00100110101101000110101101101000111111111100011010001111'. The register value is 3.
        """
        # The input column must be cast since the pyspark sha256 function only operates on binary columns.
        c = F.col(input_column).cast("binary")
        c = F.sha2(c, 256)

        # Take first 16 hex characters. The pyspark substring method is 1-indexed.
        c = F.substring(c, 1, 16)

        # Convert the hex string to binary. This operation is done as if the input column represented an integer, so
        # the resulting column might not contain 64 bits.
        c = F.conv(c, 16, 2)

        # Left pad the binary string with 0s to 64 characters, since the column might not have 64 bits.
        c = F.lpad(c, 64, "0")

        # Take the first 'precision' bits of the hash, convert to decimal, and cast to bigint.
        index = F.substring(c, 1, precision)
        index = F.conv(index, 2, 10).cast("bigint")

        # Take the final 64 - 'precision' bits of the hash. Then find the index of the first instance of '1' in the
        # column. The index starts at 1, so we do not need to modify it.
        register = F.substring(c, precision + 1, 64)
        register = F.instr(register, "1").cast("bigint")

        return [index, register]

    def _approx_count_distinct_partial_aggregation_helper(col: str) -> List[Column]:
        return _add_index_and_register_column(col)

    return _approx_count_distinct_partial_aggregation_helper


def _mean_full_aggregation(cols: List[str], window: WindowSpec) -> Column:
    # Window aggregation doesn't work with more than one built-in function like this
    #   sum(mean_clicked * count_clicked) / sum(count_clicked)
    # And it does not support UDFs on bounded windows (the kind we use)
    #   https://issues.apache.org/jira/browse/SPARK-22239
    # We work around this limitations by calculating ratio over two window aggregations
    mean_col, count_col = cols
    return functions.sum(functions.col(mean_col) * functions.col(count_col)).over(window) / functions.sum(
        count_col
    ).over(window)


# Partial aggregator used by first distinct N.
def FirstNAgg(timestamp: str, col: str, n: int, col_datatype: DataType) -> Column:
    sc = SparkContext._active_spark_context
    udf_name = f"tecton_{repr(col_datatype).lower()}_first_distinct_{n}_partial_aggregation"
    if col_datatype == StringType():
        get_or_create_udf_jar_class_loader().load_class(
            sc, "com.tecton.udfs.spark3.FirstNRegister"
        ).newInstance().register(n, udf_name, True)
    elif col_datatype == Int64Type():
        get_or_create_udf_jar_class_loader().load_class(
            sc, "com.tecton.udfs.spark3.FirstDistinctNRegisterForLongType"
        ).newInstance().register(n, udf_name)
    else:
        msg = f"Unsupported data type {col_datatype} for column {col} for first/last distinct aggregation."
        raise ValueError(msg)
    return expr(f"{udf_name}({timestamp},{col}).values")


def _make_first_distinct_n_partial_aggregation(time_key: str, n: int, col_datatype: DataType) -> Callable:
    def _first_distinct_n_partial_aggregation(cols: List[str]) -> List[Column]:
        return [FirstNAgg(time_key, cols[0], n, col_datatype)]

    return _first_distinct_n_partial_aggregation


# Partial aggregator used by last distinct N.
def LastDistinctNAgg(col1: str, col2: str, n: int, col_datatype: DataType) -> Column:
    sc = SparkContext._active_spark_context
    udf_name = f"tecton_{repr(col_datatype).lower()}_last_distinct_{n}_partial_aggregation"
    if col_datatype == StringType():
        get_or_create_udf_jar_class_loader().load_class(
            sc, "com.tecton.udfs.spark3.LastNRegister"
        ).newInstance().register(n, udf_name, True)
    elif col_datatype == Int64Type():
        get_or_create_udf_jar_class_loader().load_class(
            sc, "com.tecton.udfs.spark3.LastDistinctNRegisterForLongType"
        ).newInstance().register(n, udf_name)
    else:
        msg = f"Unsupported data type {col_datatype} for column {col1} for first/last distinct aggregation."
        raise ValueError(msg)
    return expr(f"{udf_name}({col1}, {col2}).values")


def _make_last_distinct_n_partial_aggregation(
    time_key: str, n: int, col_datatype: DataType
) -> Callable[[List[str]], List[Column]]:
    def _last_distinct_n_partial_aggregation(cols: List[str]) -> List[Column]:
        return [LastDistinctNAgg(time_key, cols[0], n, col_datatype)]

    return _last_distinct_n_partial_aggregation


def _make_first_and_last_n_continuous_partial() -> Callable[[List[str]], List[Column]]:
    def _first_and_last_n_continuous_partial(cols: List[str]) -> List[Column]:
        return [F.array(F.col(cols[0]))]

    return _first_and_last_n_continuous_partial


# Full aggregator used by both last and first distinct N.
def LimitedListConcatAgg(col1: str, n: int, keep_last_items: bool, col_datatype: Optional[DataType]) -> Column:
    sc = SparkContext._active_spark_context
    udf_name = (
        f"tecton_{repr(col_datatype).lower()}_last_distinct_{n}_full_aggregation"
        if keep_last_items
        else f"tecton_{repr(col_datatype).lower()}_first_distinct_{n}_full_aggregation"
    )

    if col_datatype == StringType():
        get_or_create_udf_jar_class_loader().load_class(
            sc, "com.tecton.udfs.spark3.LimitedListConcatRegister"
        ).newInstance().register(n, udf_name, keep_last_items)
    elif col_datatype == Int64Type():
        get_or_create_udf_jar_class_loader().load_class(
            sc, "com.tecton.udfs.spark3.LimitedListConcatRegisterForLongType"
        ).newInstance().register(n, udf_name, keep_last_items)
    else:
        msg = f"Unsupported data type {col_datatype} for column {col1} for first/last distinct aggregation."
        raise ValueError(msg)
    return expr(f"{udf_name}({col1})")


def _make_fixed_size_n_full_aggregation(
    n: int, keep_last_items: bool, col_datatype: Optional[DataType]
) -> Callable[[List[str], WindowSpec], Column]:
    def _fixed_size_n_full_aggregation(column_name: List[str], window: WindowSpec) -> Column:
        col = (LimitedListConcatAgg(column_name[0], n, keep_last_items, col_datatype).over(window)).values
        return col

    return _fixed_size_n_full_aggregation


# Full aggregator used by last non-distinct N.
def _make_last_non_distinct_n_full_aggregation(n: int) -> Callable[[List[str], WindowSpec], Column]:
    def _last_non_distinct_n_full_aggregation(cols: List[str], window: WindowSpec) -> Column:
        return functions.reverse(
            functions.slice(functions.reverse(functions.flatten(functions.collect_list(cols[0]).over(window))), 1, n)
        )

    return _last_non_distinct_n_full_aggregation


# Full aggregator used by first non-distinct N.
def _make_first_non_distinct_n_full_aggregation(n: int) -> Callable[[List[str], WindowSpec], Column]:
    def _first_non_distinct_n_full_aggregation(cols: List[Column], window: WindowSpec) -> Column:
        return functions.slice(functions.flatten(functions.collect_list(cols[0]).over(window)), 1, n)

    return _first_non_distinct_n_full_aggregation


# Partial aggregator used by last non-distinct N.
def _make_last_non_distinct_n_partial_aggregation(time_key: str, n: int) -> Callable[[List[str]], List[Column]]:
    def last_non_distinct_n_partial_aggregation(cols: List[str]) -> List[Column]:
        # Sort items in descending order based on timestamp.
        col = cols[0]
        sort_function = f"(left, right) -> case when left.{time_key} < right.{time_key} then 1 when left.{time_key} > right.{time_key} then -1 else 0 end)"
        return [
            functions.reverse(
                functions.slice(
                    functions.expr(f"array_sort(collect_list(struct({col}, {time_key})), {sort_function}"),
                    1,
                    n,
                ).getItem(col)
            )
        ]

    return last_non_distinct_n_partial_aggregation


# Partial aggregator used by last N.
def _make_last_partial_aggregation(time_key: str) -> Callable[[List[str]], List[Column]]:
    def last_partial_aggregation(cols: List[str]) -> List[Column]:
        # Sort items in descending order based on timestamp.
        col = cols[0]
        filter_condition = f" x -> x.{col} IS NOT NULL"
        sort_function = f"(left, right) -> case when left.{time_key} < right.{time_key} then 1 when left.{time_key} > right.{time_key} then -1 else 0 end)"
        return [
            functions.expr(
                f"filter(array_sort(collect_list(struct({col}, {time_key})), {sort_function},{filter_condition})"
            )
            .getItem(0)
            .getItem(col)
        ]

    return last_partial_aggregation


# Partial aggregator used by first non-distinct N.
def _make_first_non_distinct_n_partial_aggregation(time_key: str, n: int) -> Callable[[List[str]], List[Column]]:
    def first_non_distinct_n_partial_aggregation(cols: List[str]) -> List[Column]:
        # Sort items in ascending order based on timestamp.
        col = cols[0]
        sort_function = f"(left,right) -> case when left.{time_key} < right.{time_key} then -1 when left.{time_key} > right.{time_key} then 1 else 0 end)"
        return [
            functions.slice(
                functions.expr(f"array_sort(collect_list(struct({col}, {time_key})), {sort_function}"),
                1,
                n,
            ).getItem(col)
        ]

    return first_non_distinct_n_partial_aggregation


def _sum_with_default(columns: List[str], window: WindowSpec) -> Column:
    col = functions.sum(columns[0]).over(window)
    # Fill null
    col = functions.when(col.isNull(), functions.lit(0)).otherwise(col)
    return col


# population variation equation: Σ(x^2)/n - μ^2
def _var_pop_full_aggregation(cols: List[str], window: WindowSpec) -> Column:
    sum_of_squares_col, count_col, sum_col = cols
    return functions.sum(sum_of_squares_col).over(window) / functions.sum(count_col).over(window) - functions.pow(
        functions.sum(sum_col).over(window) / functions.sum(count_col).over(window), 2
    )


def _stddev_pop_full_aggregation(cols: List[str], window: WindowSpec) -> Column:
    return functions.sqrt(_var_pop_full_aggregation(cols, window))


# sample variation equation: (Σ(x^2) - (Σ(x)^2)/N)/N-1
def _var_samp_full_aggregation(cols: List[str], window: WindowSpec) -> Column:
    sum_of_squares_col, count_col, sum_col = cols
    total_count_col = functions.sum(count_col).over(window)
    # check if count is equal to 0 for divide by 0 errors
    var_samp_col = functions.when(
        total_count_col != 1,
        (
            functions.sum(sum_of_squares_col).over(window)
            - functions.pow(functions.sum(sum_col).over(window), 2) / total_count_col
        )
        / (total_count_col - functions.lit(1)),
    )
    return var_samp_col


def _stddev_samp_full_aggregation(cols: List[str], window: WindowSpec) -> Column:
    return functions.sqrt(_var_samp_full_aggregation(cols, window))


def _stddev_var_partial(col: List[str]) -> List[Column]:
    return [
        functions.sum(functions.pow(col, 2)),
        functions.count(col),
        functions.sum(functions.col(col)),
    ]


def _make_stddev_var_continuous_partial() -> Callable[[List[str]], List[Column]]:
    def _stddev_var_continuous_partial(cols: List[str]) -> List[Column]:
        col = cols[0]
        return [
            F.pow(col, 2),
            *_count_continuous_partial_aggregation([col]),
            F.col(col),
        ]

    return _stddev_var_continuous_partial


def _make_approx_count_distinct_continuous_partial() -> Callable[[List[str]], List[Column]]:
    def _approx_count_distinct_continuous_partial(cols: List[str]) -> List[Column]:
        index, register = cols[0], cols[1]
        return [F.array(index), F.array(register)]

    return _approx_count_distinct_continuous_partial


def _count_continuous_partial_aggregation(cols: List[str]) -> List[Column]:
    return [F.when(F.col(cols[0]).isNull(), F.lit(0).cast("long")).otherwise(F.lit(1).cast("long"))]


def _mean_continuous_partial_aggregation(cols: List[str]) -> List[Column]:
    return [F.col(cols[0]).cast("double"), *_count_continuous_partial_aggregation([cols[0]])]


_approx_count_distinct_intermediate_column_prefixes = ["approx_count_distinct_index", "approx_count_distinct_register"]

AGGREGATION_PLANS = {
    afpb.AGGREGATION_FUNCTION_SUM: _simple_aggregation_plan(afpb.AGGREGATION_FUNCTION_SUM, functions.sum),
    afpb.AGGREGATION_FUNCTION_MIN: _simple_aggregation_plan(afpb.AGGREGATION_FUNCTION_MIN, functions.min),
    afpb.AGGREGATION_FUNCTION_MAX: _simple_aggregation_plan(afpb.AGGREGATION_FUNCTION_MAX, functions.max),
    afpb.AGGREGATION_FUNCTION_LAST: lambda time_key, params, is_continuous, _: AggregationPlan(
        partial_aggregation_transform=_make_last_partial_aggregation(time_key),
        continuous_partial_aggregation_transform=_simple_continuous_partial_aggregation_transform(),
        full_aggregation_transform=_simple_full_aggregation_transform(
            lambda col: functions.last(col, ignorenulls=True)
        ),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(afpb.AGGREGATION_FUNCTION_LAST),
    ),
    # Needs to use COUNT for partial and SUM for full aggregation
    afpb.AGGREGATION_FUNCTION_COUNT: AggregationPlan(
        partial_aggregation_transform=_simple_partial_aggregation_transform(functions.count),
        continuous_partial_aggregation_transform=_count_continuous_partial_aggregation,
        full_aggregation_transform=_sum_with_default,
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(afpb.AGGREGATION_FUNCTION_COUNT),
    ),
    afpb.AGGREGATION_FUNCTION_MEAN: AggregationPlan(
        partial_aggregation_transform=lambda cols: [functions.mean(cols[0]), functions.count(cols[0])],
        continuous_partial_aggregation_transform=_mean_continuous_partial_aggregation,
        full_aggregation_transform=_mean_full_aggregation,
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(afpb.AGGREGATION_FUNCTION_MEAN),
    ),
    afpb.AGGREGATION_FUNCTION_LAST_DISTINCT_N: lambda time_key, params, is_continuous, col_datatype: AggregationPlan(
        partial_aggregation_transform=_make_last_distinct_n_partial_aggregation(
            time_key, params.last_n.n, col_datatype
        ),
        continuous_partial_aggregation_transform=_make_first_and_last_n_continuous_partial(),
        full_aggregation_transform=_make_fixed_size_n_full_aggregation(params.last_n.n, True, col_datatype),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_LAST_DISTINCT_N, function_params=params, is_continuous=is_continuous
        ),
    ),
    afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N: lambda time_key, params, is_continuous, _: AggregationPlan(
        partial_aggregation_transform=_make_last_non_distinct_n_partial_aggregation(time_key, params.last_n.n),
        continuous_partial_aggregation_transform=_make_first_and_last_n_continuous_partial(),
        full_aggregation_transform=_make_last_non_distinct_n_full_aggregation(params.last_n.n),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N, function_params=params, is_continuous=is_continuous
        ),
    ),
    afpb.AGGREGATION_FUNCTION_VAR_POP: AggregationPlan(
        partial_aggregation_transform=lambda cols: _stddev_var_partial(cols[0]),
        continuous_partial_aggregation_transform=_make_stddev_var_continuous_partial(),
        full_aggregation_transform=_var_pop_full_aggregation,
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(afpb.AGGREGATION_FUNCTION_VAR_POP),
    ),
    afpb.AGGREGATION_FUNCTION_STDDEV_POP: AggregationPlan(
        partial_aggregation_transform=lambda cols: _stddev_var_partial(cols[0]),
        continuous_partial_aggregation_transform=_make_stddev_var_continuous_partial(),
        full_aggregation_transform=_stddev_pop_full_aggregation,
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_STDDEV_POP
        ),
    ),
    afpb.AGGREGATION_FUNCTION_VAR_SAMP: AggregationPlan(
        partial_aggregation_transform=lambda cols: _stddev_var_partial(cols[0]),
        continuous_partial_aggregation_transform=_make_stddev_var_continuous_partial(),
        full_aggregation_transform=_var_samp_full_aggregation,
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_VAR_SAMP
        ),
    ),
    afpb.AGGREGATION_FUNCTION_STDDEV_SAMP: AggregationPlan(
        partial_aggregation_transform=lambda cols: _stddev_var_partial(cols[0]),
        continuous_partial_aggregation_transform=_make_stddev_var_continuous_partial(),
        full_aggregation_transform=_stddev_samp_full_aggregation,
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_STDDEV_SAMP
        ),
    ),
    afpb.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT: lambda time_key, params, _, __: AggregationPlan(
        partial_aggregation_transform=_make_approx_count_distinct_partial_aggregation(),
        continuous_partial_aggregation_transform=_make_approx_count_distinct_continuous_partial(),
        full_aggregation_transform=_make_approx_count_distinct_full_aggregation(params.approx_count_distinct.precision),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT
        ),
        partial_aggregation_preprocessor=_make_approx_count_distinct_partial_aggregation_helper(
            params.approx_count_distinct.precision
        ),
        intermediate_column_prefixes=_approx_count_distinct_intermediate_column_prefixes,
    ),
    afpb.AGGREGATION_FUNCTION_APPROX_PERCENTILE: lambda time_key, params, _, __: AggregationPlan(
        partial_aggregation_transform=_make_approx_percentile_partial_aggregation(params.approx_percentile.precision),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_APPROX_PERCENTILE
        ),
        continuous_partial_aggregation_transform=_make_approx_percentile_continuous_partial(),
        full_aggregation_transform=_make_approx_percentile_full_aggregation(
            params.approx_percentile.percentile, params.approx_percentile.precision
        ),
    ),
    afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N: lambda time_key, params, is_continuous, _: AggregationPlan(
        partial_aggregation_transform=_make_first_non_distinct_n_partial_aggregation(time_key, params.first_n.n),
        continuous_partial_aggregation_transform=_make_first_and_last_n_continuous_partial(),
        full_aggregation_transform=_make_first_non_distinct_n_full_aggregation(params.first_n.n),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N, function_params=params, is_continuous=is_continuous
        ),
    ),
    afpb.AGGREGATION_FUNCTION_FIRST_DISTINCT_N: lambda time_key, params, is_continuous, col_datatype: AggregationPlan(
        partial_aggregation_transform=_make_first_distinct_n_partial_aggregation(
            time_key, params.first_n.n, col_datatype
        ),
        continuous_partial_aggregation_transform=_make_first_and_last_n_continuous_partial(),
        full_aggregation_transform=_make_fixed_size_n_full_aggregation(params.first_n.n, False, col_datatype),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_FIRST_DISTINCT_N, function_params=params, is_continuous=is_continuous
        ),
    ),
}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/databricks_api/clusters__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.spark_common import clusters__client_pb2 as tecton__proto_dot_spark__common_dot_clusters____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2tecton_proto/databricks_api/clusters__client.proto\x12\x1btecton_proto.databricks_api\x1a\x30tecton_proto/spark_common/clusters__client.proto\"3\n\x12\x43lustersGetRequest\x12\x1d\n\ncluster_id\x18\x01 \x01(\tR\tclusterId\"\xc7\n\n\x14\x43lusterCreateRequest\x12_\n\nspark_conf\x18\x01 \x03(\x0b\x32@.tecton_proto.databricks_api.ClusterCreateRequest.SparkConfEntryR\tsparkConf\x12-\n\x13\x64river_node_type_id\x18\x02 \x01(\tR\x10\x64riverNodeTypeId\x12 \n\x0cnode_type_id\x18\x03 \x01(\tR\nnodeTypeId\x12\x1f\n\x0bnum_workers\x18\x04 \x01(\x05R\nnumWorkers\x12!\n\x0c\x63luster_name\x18\x05 \x01(\tR\x0b\x63lusterName\x12#\n\rspark_version\x18\x06 \x01(\tR\x0csparkVersion\x12O\n\x0e\x61ws_attributes\x18\x07 \x01(\x0b\x32(.tecton_proto.spark_common.AwsAttributesR\rawsAttributes\x12+\n\x11idempotency_token\x18\x08 \x01(\tR\x10idempotencyToken\x12i\n\x0espark_env_vars\x18\t \x03(\x0b\x32\x43.tecton_proto.databricks_api.ClusterCreateRequest.SparkEnvVarsEntryR\x0csparkEnvVars\x12\x62\n\x0b\x63ustom_tags\x18\n \x03(\x0b\x32\x41.tecton_proto.databricks_api.ClusterCreateRequest.CustomTagsEntryR\ncustomTags\x12\x37\n\x17\x61utotermination_minutes\x18\x0b \x01(\x05R\x16\x61utoterminationMinutes\x12.\n\x13\x65nable_elastic_disk\x18\x0c \x01(\x08R\x11\x65nableElasticDisk\x12K\n\tautoscale\x18\r \x01(\x0b\x32-.tecton_proto.databricks_api.ClusterAutoScaleR\tautoscale\x12N\n\x0cinit_scripts\x18\x0e \x03(\x0b\x32+.tecton_proto.spark_common.ResourceLocationR\x0binitScripts\x12\x1b\n\tpolicy_id\x18\x0f \x01(\tR\x08policyId\x12O\n\x0egcp_attributes\x18\x10 \x01(\x0b\x32(.tecton_proto.spark_common.GCPAttributesR\rgcpAttributes\x12,\n\x12\x64\x61ta_security_mode\x18\x11 \x01(\tR\x10\x64\x61taSecurityMode\x12(\n\x10single_user_name\x18\x12 \x01(\tR\x0esingleUserName\x12=\n\x1b\x61pply_policy_default_values\x18\x13 \x01(\x08R\x18\x61pplyPolicyDefaultValues\x1a<\n\x0eSparkConfEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a?\n\x11SparkEnvVarsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a=\n\x0f\x43ustomTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"T\n\x10\x43lusterAutoScale\x12\x1f\n\x0bmin_workers\x18\x01 \x01(\x05R\nminWorkers\x12\x1f\n\x0bmax_workers\x18\x02 \x01(\x05R\nmaxWorkers\"6\n\x15\x43lusterCreateResponse\x12\x1d\n\ncluster_id\x18\x01 \x01(\tR\tclusterId\"8\n\x17\x43lusterTerminateRequest\x12\x1d\n\ncluster_id\x18\x01 \x01(\tR\tclusterId\"\xf9\x01\n\x13\x43lustersGetResponse\x12\x1d\n\ncluster_id\x18\x01 \x01(\tR\tclusterId\x12#\n\rstate_message\x18\x02 \x01(\tR\x0cstateMessage\x12]\n\x12termination_reason\x18\x03 \x01(\x0b\x32..tecton_proto.databricks_api.TerminationReasonR\x11terminationReason\x12?\n\x05state\x18\x04 \x01(\x0e\x32).tecton_proto.databricks_api.ClusterStateR\x05state\"\x97\x01\n\x11TerminationReason\x12@\n\x04\x63ode\x18\x01 \x01(\x0e\x32,.tecton_proto.databricks_api.TerminationCodeR\x04\x63ode\x12@\n\x04type\x18\x02 \x01(\x0e\x32,.tecton_proto.databricks_api.TerminationTypeR\x04type\"W\n\x13\x43lusterListResponse\x12@\n\x08\x63lusters\x18\x01 \x03(\x0b\x32$.tecton_proto.databricks_api.ClusterR\x08\x63lusters\"\xc7\x02\n\x07\x43luster\x12\x1d\n\ncluster_id\x18\x01 \x01(\tR\tclusterId\x12!\n\x0c\x63luster_name\x18\x02 \x01(\tR\x0b\x63lusterName\x12#\n\rspark_version\x18\x03 \x01(\tR\x0csparkVersion\x12?\n\x05state\x18\x04 \x01(\x0e\x32).tecton_proto.databricks_api.ClusterStateR\x05state\x12U\n\x0b\x63ustom_tags\x18\x05 \x03(\x0b\x32\x34.tecton_proto.databricks_api.Cluster.CustomTagsEntryR\ncustomTags\x1a=\n\x0f\x43ustomTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01*\x7f\n\x0c\x43lusterState\x12\x0b\n\x07PENDING\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\x0e\n\nRESTARTING\x10\x02\x12\x0c\n\x08RESIZING\x10\x03\x12\x0f\n\x0bTERMINATING\x10\x04\x12\x0e\n\nTERMINATED\x10\x05\x12\t\n\x05\x45RROR\x10\x06\x12\x0b\n\x07UNKNOWN\x10\x07*\xf0\x04\n\x0fTerminationCode\x12\x1d\n\x19UNKNOWN_TERMINATION_STATE\x10\x00\x12\x10\n\x0cUSER_REQUEST\x10\x01\x12\x10\n\x0cJOB_FINISHED\x10\x02\x12\x0e\n\nINACTIVITY\x10\x03\x12\x1b\n\x17\x43LOUD_PROVIDER_SHUTDOWN\x10\x04\x12\x16\n\x12\x43OMMUNICATION_LOST\x10\x05\x12!\n\x1d\x43LOUD_PROVIDER_LAUNCH_FAILURE\x10\x06\x12\x19\n\x15SPARK_STARTUP_FAILURE\x10\x07\x12\x14\n\x10INVALID_ARGUMENT\x10\x08\x12\x1d\n\x19UNEXPECTED_LAUNCH_FAILURE\x10\t\x12\x12\n\x0eINTERNAL_ERROR\x10\n\x12\x0f\n\x0bSPARK_ERROR\x10\x0b\x12!\n\x1dMETASTORE_COMPONENT_UNHEALTHY\x10\x0c\x12\x1c\n\x18\x44\x42\x46S_COMPONENT_UNHEALTHY\x10\r\x12\x16\n\x12\x44RIVER_UNREACHABLE\x10\x0e\x12\x17\n\x13\x44RIVER_UNRESPONSIVE\x10\x0f\x12\x18\n\x14INSTANCE_UNREACHABLE\x10\x10\x12\x1c\n\x18\x43ONTAINER_LAUNCH_FAILURE\x10\x11\x12!\n\x1dINSTANCE_POOL_CLUSTER_FAILURE\x10\x12\x12\x14\n\x10REQUEST_REJECTED\x10\x13\x12\x17\n\x13INIT_SCRIPT_FAILURE\x10\x14\x12\x11\n\rTRIAL_EXPIRED\x10\x15\x12.\n*AWS_INSUFFICIENT_INSTANCE_CAPACITY_FAILURE\x10\x16*t\n\x0fTerminationType\x12\x1c\n\x18UNKNOWN_TERMINATION_TYPE\x10\x00\x12\x0b\n\x07SUCCESS\x10\x01\x12\x10\n\x0c\x43LIENT_ERROR\x10\x02\x12\x11\n\rSERVICE_FAULT\x10\x03\x12\x11\n\rCLOUD_FAILURE\x10\x04\x42\x1d\n\x19\x63om.tecton.databricks_apiP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.databricks_api.clusters__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.tecton.databricks_apiP\001'
  _CLUSTERCREATEREQUEST_SPARKCONFENTRY._options = None
  _CLUSTERCREATEREQUEST_SPARKCONFENTRY._serialized_options = b'8\001'
  _CLUSTERCREATEREQUEST_SPARKENVVARSENTRY._options = None
  _CLUSTERCREATEREQUEST_SPARKENVVARSENTRY._serialized_options = b'8\001'
  _CLUSTERCREATEREQUEST_CUSTOMTAGSENTRY._options = None
  _CLUSTERCREATEREQUEST_CUSTOMTAGSENTRY._serialized_options = b'8\001'
  _CLUSTER_CUSTOMTAGSENTRY._options = None
  _CLUSTER_CUSTOMTAGSENTRY._serialized_options = b'8\001'
  _CLUSTERSTATE._serialized_start=2565
  _CLUSTERSTATE._serialized_end=2692
  _TERMINATIONCODE._serialized_start=2695
  _TERMINATIONCODE._serialized_end=3319
  _TERMINATIONTYPE._serialized_start=3321
  _TERMINATIONTYPE._serialized_end=3437
  _CLUSTERSGETREQUEST._serialized_start=133
  _CLUSTERSGETREQUEST._serialized_end=184
  _CLUSTERCREATEREQUEST._serialized_start=187
  _CLUSTERCREATEREQUEST._serialized_end=1538
  _CLUSTERCREATEREQUEST_SPARKCONFENTRY._serialized_start=1350
  _CLUSTERCREATEREQUEST_SPARKCONFENTRY._serialized_end=1410
  _CLUSTERCREATEREQUEST_SPARKENVVARSENTRY._serialized_start=1412
  _CLUSTERCREATEREQUEST_SPARKENVVARSENTRY._serialized_end=1475
  _CLUSTERCREATEREQUEST_CUSTOMTAGSENTRY._serialized_start=1477
  _CLUSTERCREATEREQUEST_CUSTOMTAGSENTRY._serialized_end=1538
  _CLUSTERAUTOSCALE._serialized_start=1540
  _CLUSTERAUTOSCALE._serialized_end=1624
  _CLUSTERCREATERESPONSE._serialized_start=1626
  _CLUSTERCREATERESPONSE._serialized_end=1680
  _CLUSTERTERMINATEREQUEST._serialized_start=1682
  _CLUSTERTERMINATEREQUEST._serialized_end=1738
  _CLUSTERSGETRESPONSE._serialized_start=1741
  _CLUSTERSGETRESPONSE._serialized_end=1990
  _TERMINATIONREASON._serialized_start=1993
  _TERMINATIONREASON._serialized_end=2144
  _CLUSTERLISTRESPONSE._serialized_start=2146
  _CLUSTERLISTRESPONSE._serialized_end=2233
  _CLUSTER._serialized_start=2236
  _CLUSTER._serialized_end=2563
  _CLUSTER_CUSTOMTAGSENTRY._serialized_start=1477
  _CLUSTER_CUSTOMTAGSENTRY._serialized_end=1538
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/data_type__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n+tecton_proto/common/data_type__client.proto\x12\x13tecton_proto.common\"\xdb\x02\n\x08\x44\x61taType\x12\x35\n\x04type\x18\x01 \x01(\x0e\x32!.tecton_proto.common.DataTypeEnumR\x04type\x12K\n\x12\x61rray_element_type\x18\x02 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x10\x61rrayElementType\x12\x45\n\rstruct_fields\x18\x03 \x03(\x0b\x32 .tecton_proto.common.StructFieldR\x0cstructFields\x12?\n\x0cmap_key_type\x18\x04 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\nmapKeyType\x12\x43\n\x0emap_value_type\x18\x05 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x0cmapValueType\"]\n\x0bStructField\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12:\n\tdata_type\x18\x02 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x08\x64\x61taType*\x9d\x02\n\x0c\x44\x61taTypeEnum\x12\x15\n\x11\x44\x41TA_TYPE_UNKNOWN\x10\x00\x12\x13\n\x0f\x44\x41TA_TYPE_INT32\x10\x01\x12\x13\n\x0f\x44\x41TA_TYPE_INT64\x10\x02\x12\x15\n\x11\x44\x41TA_TYPE_FLOAT32\x10\x03\x12\x15\n\x11\x44\x41TA_TYPE_FLOAT64\x10\x04\x12\x14\n\x10\x44\x41TA_TYPE_STRING\x10\x05\x12\x12\n\x0e\x44\x41TA_TYPE_BOOL\x10\x06\x12\x17\n\x13\x44\x41TA_TYPE_TIMESTAMP\x10\x07\x12\x13\n\x0f\x44\x41TA_TYPE_ARRAY\x10\x08\x12\x14\n\x10\x44\x41TA_TYPE_STRUCT\x10\t\x12\x1d\n\x19\x44\x41TA_TYPE_TIMESTAMP_INT96\x10\n\x12\x11\n\rDATA_TYPE_MAP\x10\x0b\x42\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.data_type__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _DATATYPEENUM._serialized_start=514
  _DATATYPEENUM._serialized_end=799
  _DATATYPE._serialized_start=69
  _DATATYPE._serialized_end=416
  _STRUCTFIELD._serialized_start=418
  _STRUCTFIELD._serialized_end=511
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/saved_feature_data_frame__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2
from tecton_proto.common import spark_schema__client_pb2 as tecton__proto_dot_common_dot_spark__schema____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n8tecton_proto/data/saved_feature_data_frame__client.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a$tecton_proto/common/id__client.proto\x1a(tecton_proto/common/schema__client.proto\x1a.tecton_proto/common/spark_schema__client.proto\"\x95\x02\n\x19SavedFeatureDataFrameInfo\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x39\n\ncreated_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x1f\n\x0bis_archived\x18\x03 \x01(\x08R\nisArchived\x12\x1c\n\tworkspace\x18\x04 \x01(\tR\tworkspace\x12&\n\x0fis_data_deleted\x18\x05 \x01(\x08R\risDataDeleted\x12\x42\n\x0f\x64\x61ta_deleted_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\rdataDeletedAt\"\xda\x08\n\x15SavedFeatureDataFrame\x12T\n\x1asaved_feature_dataframe_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x17savedFeatureDataframeId\x12@\n\x04info\x18\x02 \x01(\x0b\x32,.tecton_proto.data.SavedFeatureDataFrameInfoR\x04info\x12-\n\x12\x64\x61taframe_location\x18\x03 \x01(\tR\x11\x64\x61taframeLocation\x12G\n\x12\x66\x65\x61ture_package_id\x18\x04 \x01(\x0b\x32\x17.tecton_proto.common.IdH\x00R\x10\x66\x65\x61turePackageId\x12G\n\x12\x66\x65\x61ture_service_id\x18\x05 \x01(\x0b\x32\x17.tecton_proto.common.IdH\x00R\x10\x66\x65\x61tureServiceId\x12\x32\n\x14\x66\x65\x61ture_package_name\x18\x06 \x01(\tH\x01R\x12\x66\x65\x61turePackageName\x12\x32\n\x14\x66\x65\x61ture_service_name\x18\x07 \x01(\tH\x01R\x12\x66\x65\x61tureServiceName\x12\x42\n\x1cstate_update_entry_commit_id\x18\x08 \x01(\tB\x02\x18\x01R\x18stateUpdateEntryCommitId\x12\x35\n\x15join_key_column_names\x18\t \x03(\tB\x02\x18\x01R\x12joinKeyColumnNames\x12\x36\n\x15timestamp_column_name\x18\n \x01(\tB\x02\x18\x01R\x13timestampColumnName\x12<\n\x06schema\x18\x0b \x01(\x0b\x32 .tecton_proto.common.SparkSchemaB\x02\x18\x01R\x06schema\x12\x44\n\x04type\x18\x0c \x01(\x0e\x32,.tecton_proto.data.SavedFeatureDataFrameTypeB\x02\x18\x01R\x04type\x12\x46\n\x06\x66ormat\x18\r \x01(\x0e\x32..tecton_proto.data.SavedFeatureDataFrameFormatR\x06\x66ormat\x12\x42\n\x0eunified_schema\x18\x0e \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\runifiedSchema\x12\x46\n\rsaved_dataset\x18\x0f \x01(\x0b\x32\x1f.tecton_proto.data.SavedDatasetH\x02R\x0csavedDataset\x12I\n\x0elogged_dataset\x18\x10 \x01(\x0b\x32 .tecton_proto.data.LoggedDatasetH\x02R\rloggedDatasetB\x0f\n\rsource_fco_idB\x11\n\x0fsource_fco_nameB\x06\n\x04kind\"\xb3\x01\n\x0cSavedDataset\x12\x41\n\x10\x63reation_task_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x0e\x63reationTaskId\x12,\n\x12\x65vent_column_names\x18\x02 \x03(\tR\x10\x65ventColumnNames\x12\x32\n\x15timestamp_column_name\x18\x03 \x01(\tR\x13timestampColumnName\"v\n\rLoggedDataset\x12\x31\n\x15join_key_column_names\x18\x01 \x03(\tR\x12joinKeyColumnNames\x12\x32\n\x15timestamp_column_name\x18\x02 \x01(\tR\x13timestampColumnName*?\n\x19SavedFeatureDataFrameType\x12\x0b\n\x07NOT_SET\x10\x00\x12\t\n\x05SAVED\x10\x01\x12\n\n\x06LOGGED\x10\x02*\xaf\x01\n\x1bSavedFeatureDataFrameFormat\x12/\n+SAVED_FEATURED_DATAFRAME_FORMAT_UNSPECIFIED\x10\x00\x12*\n&SAVED_FEATURED_DATAFRAME_FORMAT_LEGACY\x10\x01\x12\x33\n/SAVED_FEATURED_DATAFRAME_FORMAT_UNIFIED_STORAGE\x10\x02\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.saved_feature_data_frame__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _SAVEDFEATUREDATAFRAME.fields_by_name['state_update_entry_commit_id']._options = None
  _SAVEDFEATUREDATAFRAME.fields_by_name['state_update_entry_commit_id']._serialized_options = b'\030\001'
  _SAVEDFEATUREDATAFRAME.fields_by_name['join_key_column_names']._options = None
  _SAVEDFEATUREDATAFRAME.fields_by_name['join_key_column_names']._serialized_options = b'\030\001'
  _SAVEDFEATUREDATAFRAME.fields_by_name['timestamp_column_name']._options = None
  _SAVEDFEATUREDATAFRAME.fields_by_name['timestamp_column_name']._serialized_options = b'\030\001'
  _SAVEDFEATUREDATAFRAME.fields_by_name['schema']._options = None
  _SAVEDFEATUREDATAFRAME.fields_by_name['schema']._serialized_options = b'\030\001'
  _SAVEDFEATUREDATAFRAME.fields_by_name['type']._options = None
  _SAVEDFEATUREDATAFRAME.fields_by_name['type']._serialized_options = b'\030\001'
  _SAVEDFEATUREDATAFRAMETYPE._serialized_start=1939
  _SAVEDFEATUREDATAFRAMETYPE._serialized_end=2002
  _SAVEDFEATUREDATAFRAMEFORMAT._serialized_start=2005
  _SAVEDFEATUREDATAFRAMEFORMAT._serialized_end=2180
  _SAVEDFEATUREDATAFRAMEINFO._serialized_start=241
  _SAVEDFEATUREDATAFRAMEINFO._serialized_end=518
  _SAVEDFEATUREDATAFRAME._serialized_start=521
  _SAVEDFEATUREDATAFRAME._serialized_end=1635
  _SAVEDDATASET._serialized_start=1638
  _SAVEDDATASET._serialized_end=1817
  _LOGGEDDATASET._serialized_start=1819
  _LOGGEDDATASET._serialized_end=1937
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/canary/type__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&tecton_proto/canary/type__client.proto\x12\x13tecton_proto.canary*T\n\nCanaryType\x12\x1b\n\x17\x43\x41NARY_TYPE_UNSPECIFIED\x10\x00\x12\x14\n\x10\x43\x41NARY_TYPE_BASE\x10\x01\x12\x13\n\x0f\x43\x41NARY_TYPE_NEW\x10\x02\x42\x15\n\x11\x63om.tecton.canaryP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.canary.type__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.canaryP\001'
  _CANARYTYPE._serialized_start=63
  _CANARYTYPE._serialized_end=147
# @@protoc_insertion_point(module_scope)



import logging
from enum import Enum


logger = logging.getLogger(__name__)


class ValidationMode(str, Enum):
    EXPLICIT = "explicit"
    AUTOMATIC = "auto"
    SKIP = "skip"


def set_validation_mode(mode: ValidationMode):
    """Deprecated"""
    logger.warning(
        "The set_validation_mode() method is deprecated and will be removed in a future version. As of "
        "Tecton version 1.0, objects are automatically validated on creation"
    )



import logging

import py4j.protocol
from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from pyspark.sql.functions import lit
from pyspark.sql.types import LongType
from pyspark.sql.types import StructField
from pyspark.sql.types import StructType

from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.materialization_utils import batch_write_to_online_store
from tecton_materialization.materialization_utils import has_prior_delta_commit
from tecton_materialization.materialization_utils import set_up_online_store_sink
from tecton_materialization.materialization_utils import wait_for_metric_scrape
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_spark import feature_view_spark_utils
from tecton_spark.data_observability import create_feature_metrics_collector
from tecton_spark.materialization_plan import MATERIALIZED_RAW_DATA_END_TIME
from tecton_spark.offline_store import OfflineStoreWriterParams
from tecton_spark.offline_store import get_offline_store_writer
from tecton_spark.time_utils import convert_timestamp_to_epoch


logger = logging.getLogger(__name__)

idempotence_key = "ingestPath"


def ingest_pushed_df(spark: SparkSession, raw_df: DataFrame, materialization_task_params: MaterializationTaskParams):
    ingest_task_info = materialization_task_params.ingest_task_info
    ingest_path = ingest_task_info.ingest_parameters.ingest_path
    spark.conf.set(
        "spark.databricks.delta.commitInfo.userMetadata",
        f'{{"{idempotence_key}":"{ingest_path}"}}',
    )
    fd = feature_definition_from_task_params(materialization_task_params)

    feature_view_spark_utils.validate_df_columns_and_feature_types(raw_df, fd.view_schema)

    # Drop extra columns
    df = raw_df.select(*fd.view_schema.column_names())

    timestamp_key = fd.timestamp_key
    assert timestamp_key is not None
    version = fd.get_feature_store_format_version

    if ingest_task_info.ingest_parameters.write_to_online_feature_store:
        logger.info(f"Ingesting to the OnlineStore FT: {fd.id}")
        # Find the last timestamp
        raw_data_end_time_ts = df.agg({timestamp_key: "max"}).collect()[0][0]
        raw_data_end_time_epoch = convert_timestamp_to_epoch(raw_data_end_time_ts, version)

        online_df = df.withColumn(MATERIALIZED_RAW_DATA_END_TIME, lit(raw_data_end_time_epoch))

        sink = set_up_online_store_sink(spark, materialization_task_params)
        metrics_collector = create_feature_metrics_collector(spark, materialization_task_params)

        # collect metrics
        online_df = metrics_collector.observe(online_df)

        # TODO: For large DFs consider splitting into "chunks" to load balance across partitions and FSW instances
        batch_write_to_online_store(online_df, materialization_task_params, sink, fd.id, is_status=False)

        # Status table
        status_df = [[raw_data_end_time_epoch]]
        schema = StructType([StructField(MATERIALIZED_RAW_DATA_END_TIME, LongType())])
        status_df = spark.createDataFrame(status_df, schema=schema)
        batch_write_to_online_store(status_df, materialization_task_params, sink, fd.id, is_status=True)
        if sink is not None:
            # Temporarily wrap closeGlobalResources() for ingest since this
            # runs in the internal cluster. This is because we may run this
            # code before the new OSW jar is installed on the internal cluster.
            # This isn't a problem for other types of materialization since
            # they use short-lived clusters that use materialization library
            # and OSW from the same release.
            #
            # TODO(brian): remove try/except after all validation clusters are
            # restarted.
            try:
                sink.closeGlobalResources()
            except py4j.protocol.Py4JError:
                pass

        # publish metrics
        try:
            metrics_collector.publish()
        except Exception as e:
            logger.error(f"Metrics publishing failed: {e}")

    # skip delta write if we detect a prior job has successfully committed the data already
    if ingest_task_info.ingest_parameters.write_to_offline_feature_store and not has_prior_delta_commit(
        spark, materialization_task_params, idempotence_key, ingest_path
    ):
        logger.info(f"Ingesting to the OfflineStore FT: {fd.id}")
        offline_store_params = OfflineStoreWriterParams(
            s3_path=materialization_task_params.offline_store_path,
            always_store_anchor_column=False,
            time_column=timestamp_key,
            join_key_columns=fd.join_keys,
            is_continuous=False,
        )
        offline_store_config = fd.offline_store_config
        assert offline_store_config.HasField("delta"), "FeatureTables do not support Parquet-based Offline storage"
        store_writer = get_offline_store_writer(offline_store_params, fd, version, spark)
        store_writer.upsert_dataframe(df)

    wait_for_metric_scrape()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/feature_analytics/feature_analytics__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import fco_locator__client_pb2 as tecton__proto_dot_common_dot_fco__locator____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n>tecton_proto/feature_analytics/feature_analytics__client.proto\x12\x1etecton_proto.feature_analytics\x1a\x1fgoogle/protobuf/timestamp.proto\x1a-tecton_proto/common/fco_locator__client.proto\x1a$tecton_proto/common/id__client.proto\"\x8e\x03\n\x19\x46\x65\x61tureViewSimilarityPair\x12K\n\x11\x66v_locator_second\x18\x01 \x01(\x0b\x32\x1f.tecton_proto.common.FcoLocatorR\x0f\x66vLocatorSecond\x12R\n\x19workspace_state_id_second\x18\x02 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x16workspaceStateIdSecond\x12\x1e\n\nsimilarity\x18\x03 \x01(\x02R\nsimilarity\x12M\n\x04type\x18\x04 \x01(\x0e\x32\x39.tecton_proto.feature_analytics.FeatureViewSimilarityTypeR\x04type\x12*\n\x11\x66v_details_second\x18\x05 \x01(\tR\x0f\x66vDetailsSecond\x12\x35\n\x16similarity_description\x18\x06 \x01(\tR\x15similarityDescription\"\xb4\x02\n\x1a\x46\x65\x61tureViewSimilarityGroup\x12I\n\x10\x66v_locator_first\x18\x01 \x01(\x0b\x32\x1f.tecton_proto.common.FcoLocatorR\x0e\x66vLocatorFirst\x12P\n\x18workspace_state_id_first\x18\x02 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x15workspaceStateIdFirst\x12(\n\x10\x66v_details_first\x18\x03 \x01(\tR\x0e\x66vDetailsFirst\x12O\n\x05pairs\x18\x04 \x03(\x0b\x32\x39.tecton_proto.feature_analytics.FeatureViewSimilarityPairR\x05pairs\"\xbb\x01\n\x1f\x46\x65\x61tureSimilarityAnalysisResult\x12?\n\ranalysis_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x61nalysisTime\x12W\n\tfv_groups\x18\x02 \x03(\x0b\x32:.tecton_proto.feature_analytics.FeatureViewSimilarityGroupR\x08\x66vGroups*\x93\x01\n\x19\x46\x65\x61tureViewSimilarityType\x12#\n\x1f\x46\x45\x41TURE_VIEW_SIMILARITY_UNKNOWN\x10\x00\x12%\n!FEATURE_VIEW_SIMILARITY_DUPLICATE\x10\x01\x12*\n&FEATURE_VIEW_SIMILARITY_NEAR_DUPLICATE\x10\x02\x42 \n\x1c\x63om.tecton.feature_analyticsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.feature_analytics.feature_analytics__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034com.tecton.feature_analyticsP\001'
  _FEATUREVIEWSIMILARITYTYPE._serialized_start=1119
  _FEATUREVIEWSIMILARITYTYPE._serialized_end=1266
  _FEATUREVIEWSIMILARITYPAIR._serialized_start=217
  _FEATUREVIEWSIMILARITYPAIR._serialized_end=615
  _FEATUREVIEWSIMILARITYGROUP._serialized_start=618
  _FEATUREVIEWSIMILARITYGROUP._serialized_end=926
  _FEATURESIMILARITYANALYSISRESULT._serialized_start=929
  _FEATURESIMILARITYANALYSISRESULT._serialized_end=1116
# @@protoc_insertion_point(module_scope)
